import logging
import numpy as np
from .subgraphcollection import SubgraphCollection
from offsetbasedgraph import Interval, IntervalCollection, BlockArray
from .legacy.sparsepileup import SparseAreasDict, intervals_to_start_and_ends
from .postprocess.holecleaner import HolesCleaner
from .sparsepileupv2 import RpScore


class DensePileupData:

    def __init__(self, graph, ndim=1, base_value=0, dtype=None, padding=0):
        self._padding = padding
        self._values = None
        self._node_indexes = None
        self._graph = graph
        self.min_node = None
        self._touched_nodes = set()
        self.ndim = ndim
        self.dtype = dtype
        self._create_empty(ndim, base_value=base_value)

    def _create_empty(self, ndim=1, base_value=0):
        logging.info("Sorting nodes")
        #if isinstance(self._graph.blocks, BlockArray):
        #    self._nodes = self._graph.blocks.get_sorted_node_ids()
        #    logging.info("Used direct method to get sorted nodes")
        #else:
        self._nodes = sorted(self._graph.blocks.keys())

        sorted_nodes = self._nodes
        self.min_node = sorted_nodes[0]
        max_node = sorted_nodes[-1]
        span = max_node - self.min_node + 1
        logging.info("Counting basepairs")
        n_elements = self._graph.number_of_basepairs()

        if self.dtype is not None:
            self._values = np.zeros(n_elements+self._padding)
        else:
            self._values = np.zeros(n_elements+self._padding)

        if base_value > 0:
            self._values += base_value

        if isinstance(self._graph.blocks, BlockArray):
            # Quicker way to make node_indexes array
            logging.info("Using quick way to init densepileup (using cumsum on np block array)")
            self._node_indexes = np.cumsum(self._graph.blocks._array, dtype=np.uint32)
            logging.info("Node indexes created...")
        else:
            print(span)
            self._node_indexes = np.zeros(span+1, dtype=np.uint32)
            offset = 0
            for i, node in enumerate(self._nodes):
                index = node - self.min_node
                self._node_indexes[index] = offset
                offset += self.node_size(node)
            self._node_indexes[-1] = offset
        logging.info("Dense pileup inited")

    def sum(self):
        return np.sum(self._values)

    def sanitize_node(self):
        return

    def node_size(self, node):
        return self._graph.node_size(node)

    def values(self, tmp_node):
        node = abs(tmp_node)
        sign = tmp_node//node
        index = node - self.min_node
        start = self._node_indexes[index]
        end = start + self.node_size(node)
        vals = self._values[start:end]
        if sign < 0:
            return vals[::-1]
        return vals

    def values_in_range(self, node, start, end):
        index = node - self.min_node
        array_start = self._node_indexes[index] + start
        array_end = self._node_indexes[index] + end
        return self._values[array_start:array_end]

    def node_range_to_value_indexes(self, node, start, end):
        if node < 0:
            node = -node
            new_start = self.node_size(node) - end
            end = self.node_size(node) - start
            start = new_start

        node_size = self.node_size(node)
        start = min(node_size, max(start, 0))
        end = max(0, min(node_size, end))

        index = node - self.min_node
        array_start = self._node_indexes[index] + start
        array_end = self._node_indexes[index] + end
        return array_start, array_end

    def set_values(self, node, start, end, value):
        index = node - self.min_node
        array_start = self._node_indexes[index] + start
        array_end = self._node_indexes[index] + end
        self._values[array_start:array_end] = value
        self._touched_nodes.add(node)

    def add_value(self, node, start, end, value):
        index = node - self.min_node
        array_start = self._node_indexes[index] + start
        array_end = self._node_indexes[index] + end
        self._values[array_start:array_end] += value
        self._touched_nodes.add(node)

    def add_value_to_full_node(self, node, value):
        index = node - self.min_node
        array_start = self._node_indexes[index]
        array_end = array_start + self.node_size(node)
        self._values[array_start:array_end] += value
        self._touched_nodes.add(node)

    def set_full_node_value(self, node, value):
        index = node - self.min_node
        array_start = self._node_indexes[index]
        array_end = array_start + self.node_size(node)
        self._values[array_start:array_end] = value
        self._touched_nodes.add(node)

    def get_subset_max_value(self, node_id, start, end):
        return np.max(self.values(node_id)[start:end])

    def get_subset_sum(self, node_id, start, end):
        return np.sum(self.values(node_id)[start:end])

    def score(self, node_id, start, end):
        return RpScore(self.get_subset_max_value(node_id, start, end),
                       self.get_subset_sum(node_id, start, end))

    def scale(self, factor):
        self._values *= factor

    def set_new_values(self, values):
        self._values = values

    def fill_existing_hole(self, node, start, end, value):
        assert np.all(self.values_in_range(node, start, end) == 0)
        self.set_values(node, start, end, value)

    def get_sparse_indexes_and_values(self, node):
        values = self.values(node)
        diffs = np.ediff1d(values, to_begin=np.array([1]))
        indexes = np.where(diffs != 0)
        values = values[indexes]
        indexes = np.append(indexes, self.node_size(node))
        return indexes, values

    def find_valued_areas(self, node, value, changes=None):
        # Return list[start, end, start2, end2,...] having this value inside
        index = node - self.min_node
        start = self._node_indexes[index]
        end = start + self.node_size(node)
        changes = np.nonzero(changes[start:end])[0]+1
        if changes.size and changes[-1] == end-start:
            changes = changes[:-1]
        # is_value = values == value
        # changes = np.nonzero(np.ediff1d(is_value))[0]+1
        if self._values[start] == value:
            if self._values[end-1] == value:
                return [0] + list(changes)+[end-start]
            return [0] + list(changes)
        if self._values[end-1]:
            return list(changes)+[end-start]
        return list(changes)

    def nodes(self):
        return self._touched_nodes

    def copy(self):
        new = DensePileupData(self._graph)
        new._values = self._values
        new._touched_nodes = self._touched_nodes
        return new

    def threshold_copy(self, cutoff):
        new = self.copy()
        new._values = new._values >= cutoff
        logging.info("Thresholding done.")
        return new

    def threshold(self, cutoff):
        self._values = self._values >= cutoff

    def index_value_pairs(self, node):
        indexes, values = self.get_sparse_indexes_and_values(node)
        assert len(indexes) >= 2
        assert len(values) >= 1
        lines = list(zip(
            indexes[:-1],
            indexes[1:],
            values
            ))
        assert len(lines) >= 1
        return lines

    def get_bed_graph_lines(self):
        for node in self.nodes():
            lines = self.index_value_pairs(node)
            for line in lines:
                yield "%s\t%s\t%s\t%s\n" % (node, line[0], line[1], line[2])

    def __len__(self):
        return len(self.nodes())

    def __eq__(self, other):
        for node in self.nodes():
            indexes, values = self.get_sparse_indexes_and_values(node)
            #print("   Checking node %d" % node)
            other_indexes, other_values = other.get_sparse_indexes_and_values(node)
            #print("   Values: %s / %s" % (values, other_values))
            print(indexes, other_indexes)
            if not np.all(indexes == other_indexes):
                print("Indices %s != %s" % (indexes, other_indexes))
                return False

            print(values, other_values)
            if np.any(np.abs(values -  other_values) > 1e-5):
                print("Values %s != %s" % (values, other_values))
                print()
                return False

        return True

    def get_interval_values(self, interval):

        values = np.zeros(interval.length())
        offset = 0
        # find_reversed = False
        # use_interval = interval

        if all([rp < 0 for rp in interval.region_paths]):
            # Reverse before finding
            # find_reversed = True
            interval = interval.get_reverse()
        # else:
            # assert all([rp > 0 for rp in interval.region_paths]), \
            #     "This method only supports intervals with single rp direction"

        for i, rp in enumerate(interval.region_paths):
            assert rp > 0, "Currently only implemented for forward directed intervals"
            start = 0
            end = self._graph.node_size(rp)
            if i == 0:
                start = interval.start_position.offset
            if i == len(interval.region_paths) - 1:
                end = interval.end_position.offset

            values_in_rp = self.values_in_range(rp, start, end)
            values[offset:offset + (end - start)] = values_in_rp

            offset += end-start

        return values

    def value_indexes_to_nodes_and_offsets(self, indexes):
        """
        Takes indexes referring to positions in self._values
        and returns list of (node, offset).
        Indexes must be sorted
        """

        positions = []
        # Map indexes to node and offset
        length_offset = 0
        nodes = self._nodes
        current_node_index = 0
        i = 0
        while i < len(indexes):
            index = indexes[i]
            current_node = nodes[current_node_index]
            node_size = self.node_size(current_node)
            next_node_start = length_offset + node_size
            if index >= next_node_start:
                current_node_index += 1
                length_offset += node_size
                continue

            if index < next_node_start:
                positions.append((current_node, index - length_offset))

            i += 1
        return positions


class DensePileup:
    def __init__(self, graph, ndim=1, base_value=0, dtype=None):
        logging.info("Initing dense pileup")
        self.graph = graph
        self.data = DensePileupData(
            graph, ndim=ndim, base_value=base_value, dtype=dtype)

    def add_interval(self, interval):
        for i, rp in enumerate(interval.region_paths):
            start = 0
            end = self.data._graph.node_size(rp)
            if i == 0:
                start = interval.start_position.offset
            if i == len(interval.region_paths) - 1:
                end = interval.end_position.offset
            self.data.values(rp)[start:end] += 1
            self.data._touched_nodes.add(abs(rp))

    def add_areas(self, areas):
        for area, intervals in areas.items():
            self.data.values(area)[intervals[0]:intervals[1]] += 1
        self.data._touched_nodes.add(area)

    @classmethod
    def from_base_value(cls, graph, base_value):
        pileup = cls(graph, base_value=base_value)
        return pileup

    def __str__(self):
        out = "Densepileup \n"
        for node in self.data._touched_nodes:
            #out += "  Node %d: %s, %s\n" % (node, self.data.values(node), self.data.get_sparse_indexes_and_values(node))
            out += "  Node %d: %s\n" % (node, self.data.get_sparse_indexes_and_values(node))

        return out

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        return self.data == other.data

    def sum(self):
        return self.data.sum()

    def mean(self):
        graph_size = sum([self.graph.node_size(b) for b in self.graph.blocks])
        mean = self.sum() / graph_size
        return mean

    def scale(self, factor):
        self.data.scale(factor)

    def fill_small_wholes_on_dag(self, max_size):
        from .dagholecleaner import DagHoleCleaner
        logging.info("Cleaning holes using Dag Hole Cleaner")
        cleaner = DagHoleCleaner(self, max_size)
        cleaner.run()
        logging.info("Done cleaning holes")

    def set_new_values(self, values):
        self.data.set_new_values(values)

    def fill_small_wholes(self, max_size, write_holes_to_file=None, touched_nodes=None):
        cleaner = HolesCleaner(self, max_size, touched_nodes=touched_nodes)
        areas = cleaner.run()
        n_filled = 0

        hole_intervals = []
        for node_id in areas.areas:
            if touched_nodes is not None:
                if node_id not in touched_nodes:
                    continue

            starts = areas.get_starts(node_id)
            ends = areas.get_ends(node_id)
            for start, end in zip(starts, ends):
                logging.debug("Filling hole %s, %d, %d. Node size: %d" % (
                    node_id, start, end, self.graph.node_size(node_id)))

                if start == end:
                    logging.warning("Trying to fill hole of 0 length")
                    continue

                self.data.fill_existing_hole(node_id, start, end, True)

                n_filled += 1
                assert end - start <= max_size
                hole_intervals.append(Interval(start, end, [node_id]))

        logging.info(
            "Filled %d small holes (splitted into holes per node)" % n_filled)

        if write_holes_to_file is not None:
            intervals = IntervalCollection(hole_intervals)
            intervals.to_file(write_holes_to_file, text_file=True)

        self.sanitize()

    def sanitize(self):
        logging.info("Sanitizing sparse pileup")
        logging.info("Sanitize done")

    def find_valued_areas(self, value):
        logging.info("Finding valued areas equal to %d" % value)
        changes = np.diff(self.data._values == value)
        # start_values = self.data._values[
        #     self.data._node_indexes[:-1]]
        # end_values = self.data._values[
        #     self.data._node_indexes[1:]-1]
        if value:
            return SparseAreasDict({
                node_id: self.data.find_valued_areas(node_id, value, changes)
                for node_id in self.data._graph.blocks
            }, graph=self.graph)
        else:
            return SparseAreasDict(
                {node_id: self.data.find_valued_areas(node_id, value, changes)
                 for node_id in self.data._touched_nodes},
                graph=self.graph)

    @classmethod
    def from_intervals(cls, graph, intervals):
        starts, ends = intervals_to_start_and_ends(graph, intervals)
        return cls.from_starts_and_ends(graph, starts, ends)

    @classmethod
    def from_starts_and_ends(cls, graph, starts, ends, dtype=bool):
        pileup = cls(graph)
        for node in starts:
            for i, start in enumerate(starts[node]):
                end = ends[node][i]
                pileup.data.add_value(node, start, end, 1)

        return pileup

    def to_subgraphs(self):
        # Returns a list of areas which each is a subgraph
        collection = SubgraphCollection.from_pileup(self.graph, self)
        return collection

    def threshold_copy(self, cutoff):
        new_pileup = self.__class__(self.graph)
        new_pileup.data = self.data.threshold_copy(cutoff)
        return new_pileup

    def threshold(self, cutoff):
        self.data.threshold(cutoff)

    @classmethod
    def from_bed_graph(cls, graph, filename):
        raise NotImplementedError()

    @classmethod
    def from_bed_file(cls, graph, filename):
        raise NotImplementedError()

    def to_bed_file(self, filename):
        raise NotImplementedError()

    def remove_small_peaks(self, min_size):
        logging.info("Initing cleaner")
        cleaner = PeaksCleaner(self, min_size)
        logging.info("Running cleaner")
        areas = cleaner.run()
        logging.info("Removing emty areas")

        logging.info("Creating pileup using results from cleaner")
        pileup = self.from_areas_collection(self.graph, [areas])
        logging.info("Tresholding")
        pileup.threshold(0.5)
        return pileup

    def to_bed_graph(self, filename):
        f = open(filename, "w")
        i = 0
        for line in self.data.get_bed_graph_lines():
            f.write(line)
            i += 1
        f.close()
        self.filename = filename
        self.is_written = True
        return filename

    def to_bed_file(self, filename):
        f = open(filename, "w")
        areas = self.find_valued_areas(True)
        for node_id, idxs in areas.items():
            for i in range(len(idxs)//2):
                interval = (node_id, idxs[2*i], idxs[2*i+1])
                f.write("%s\t%s\t%s\t.\t.\t.\n" % interval)
        f.close()
        return filename

    def equals_old_sparse_pileup(self, old_pileup):
        # For tests to pass temporarily
        for node in self.data.nodes():
            indexes, values = self.data.get_sparse_indexes_and_values(node)
            other_indexes = old_pileup.data[node].all_idxs()
            if not np.all(indexes == other_indexes):
                return False

            other_values = old_pileup.data[node].all_values()
            if not np.allclose(values, other_values):
                return False

        return True

    @classmethod
    def create_from_old_sparsepileup(cls, old_pileup):
        # TMP method for tests to pass
        graph = old_pileup.graph
        pileup = DensePileup(graph)

        for node in old_pileup.data:

            values = old_pileup.data[node].all_values()
            indexes = old_pileup.data[node].all_idxs()

            i = 0
            for start, end in zip(indexes[0:-1], indexes[1:]):
                value = values[i]
                pileup.data.set_values(node, start, end, value)
                i += 1

        return pileup

    def set_area_to_value(self, areas, value):
        for node_id in areas.full_areas:
            self.data.set_full_node_value(node_id, value)

        for node_id, internal_intervals in areas.internal_intervals.items():
            assert node_id > 0
            self.data.set_values(node_id, internal_intervals[0], internal_intervals[1], value)

        for node_id, start in areas.starts.items():
            node_size = self.graph.node_size(node_id)
            pileup_end = start
            pileup_start = 0
            if node_id < 0:
                pileup_end = node_size
                pileup_start = node_size - start
                node_id = -node_id

            #print("   Processing start %d for node %d, adding start-end %d-%d" % (start, node_id, pileup_start, pileup_end))
            self.data.set_values(node_id, pileup_start, pileup_end, value)

    def area_is_not_empty(self, areas):
        # Returns true if area is covered by anything anywhere.
        # Areas is Binary cont areas dict
        for node_id in areas.full_areas:
            self.data.add_value_to_full_node(node_id, 1)

        for node_id, internal_intervals in areas.internal_intervals.items():
            assert node_id > 0
            self.data.add_value(node_id, internal_intervals[0], internal_intervals[1], 1)

        for node_id, start in areas.starts.items():
            node_size = self.graph.node_size(node_id)
            pileup_end = start
            pileup_start = 0
            if node_id < 0:
                pileup_end = node_size
                pileup_start = node_size - start
                node_id = -node_id

            #print("   Processing start %d for node %d, adding start-end %d-%d" % (start, node_id, pileup_start, pileup_end))
            self.data.add_value(node_id, pileup_start, pileup_end, 1)

    def add_area(self, areas):
        for node_id in areas.full_areas:
            self.data.add_value_to_full_node(node_id, 1)

        for node_id, internal_intervals in areas.internal_intervals.items():
            assert node_id > 0
            self.data.add_value(node_id, internal_intervals[0], internal_intervals[1], 1)

        for node_id, start in areas.starts.items():
            node_size = self.graph.node_size(node_id)
            pileup_end = start
            pileup_start = 0
            if node_id < 0:
                pileup_end = node_size
                pileup_start = node_size - start
                node_id = -node_id

            #print("   Processing start %d for node %d, adding start-end %d-%d" % (start, node_id, pileup_start, pileup_end))
            self.data.add_value(node_id, pileup_start, pileup_end, 1)

    @classmethod
    def create_from_binary_continous_areas(cls, graph, areas_list):
        pileup = cls(graph, dtype=np.uint8)
        i = 0
        for areas in areas_list:
            if i % 5000 == 0:
                logging.info("Processing read %d" % i)
            i += 1
            pileup.add_area(areas)

        return pileup

    @classmethod
    def from_sparse_files(cls, graph, base_file_name):
        # TODO: Use np.save(...)
        pileup = cls(graph)
        indexes = np.load(base_file_name + "_indexes.npy")
        indexes = indexes[:-1]  # Remove length added to end
        assert np.all(indexes >= 0)
        values = np.load(base_file_name + "_values.npy")
        if isinstance(values[0], np.bool_):
            logging.warning("Converting from bool to uint8 to allow reading.")
            values = values.astype(np.int8)

        diffs = np.ediff1d(values, to_begin=[values[0]])
        pileup_vals = pileup.data._values
        pileup_vals[indexes] = diffs
        pileup_vals = np.cumsum(pileup_vals)
        pileup.data._values = pileup_vals
	
        try:
            touched_nodes = np.load(
                base_file_name + "_touched_nodes.npy")
            pileup.data._touched_nodes = set(list(touched_nodes))
        except FileNotFoundError:
            logging.warning("Touched nodes not found on file. Not setting touched nodes")

        return pileup

    def to_sparse_files(self, file_base_name):
        vals = self.data._values
        assert np.all(vals >= 0)
        #vals[np.where(vals < truncate_below)] = 0
        diffs = np.ediff1d(vals, to_begin=[1])
        indexes = np.nonzero(diffs)[0]
        #indexes = np.insert(indexes, 0, 0)
        values = vals[indexes]

        # Add length to end and 0 to start
        indexes = np.append(indexes, len(self.data._values))

        np.save(file_base_name + "_touched_nodes.npy",
                   np.array(list(self.data._touched_nodes)))
        np.save(file_base_name + "_indexes.npy", indexes)
        np.save(file_base_name + "_values.npy", values)

        logging.info("Saved p values indexes/values to files")



import logging
import numpy as np
from offsetbasedgraph import IntervalCollection, DirectedInterval
from .densepileup import DensePileup
from .sampleandcontrolcreator import SampleAndControlCreator
from .sparsepvalues import PValuesFinder, PToQValuesMapper, QValuesFinder
from .postprocess import HolesCleaner, SparseMaxPaths
from .sparsediffs import SparseValues
from .reporter import Reporter
IntervalCollection.interval_class = DirectedInterval


class Configuration:
    def __init__(self, skip_filter_duplicates=False,
                 graph_is_partially_ordered=False,
                 skip_read_validation=False,
                 save_tmp_results_to_file=False,
                 p_val_cutoff=0.1,
                 use_global_min_value=None):
        self.skip_filter_duplicates = skip_filter_duplicates
        self.graph_is_partially_ordered = graph_is_partially_ordered
        self.skip_read_validation = skip_read_validation
        self.save_tmp_results_to_file = save_tmp_results_to_file
        self.p_val_cutoff = p_val_cutoff
        self.use_global_min_value = use_global_min_value

    @classmethod
    def default(cls):
        return cls()


class CallPeaks(object):

    def __init__(self, graph, out_file_base_name):
        self.graph = graph
        self.out_file_base_name = out_file_base_name

        self.sample_intervals = None
        self.control_intervals = None
        self.sample_pileup = None
        self.control_pileup = None
        self.p_values_pileup = None
        self.p_to_q_values_mapping = None
        self.q_values_pileup = None
        self.peaks_as_subgraphs = None
        self.touched_nodes = None
        self.max_path_peaks = None
        self._reporter = Reporter(self.out_file_base_name)

    def run_pre_callpeaks(self, has_control, experiment_info,
                          linear_map, configuration=None):
        if configuration is None:
            configuration = Configuration.default()
            logging.warning("Config is not set. Setting to default")
        creator = SampleAndControlCreator(
            self.graph,
            self.sample_intervals,
            self.control_intervals,
            experiment_info,
            out_file_base_name=self.out_file_base_name,
            has_control=has_control,
            linear_map=linear_map,
            configuration=configuration
            )
        creator.run()
        self.sample_pileup = creator._sample_pileup
        self.control_pileup = creator._control_pileup
        self.touched_nodes = creator.touched_nodes

    def get_p_values(self):
        assert self.sample_pileup is not None
        assert self.control_pileup is not None
        self.p_values_pileup = PValuesFinder(
            self.sample_pileup, self.control_pileup).get_p_values_pileup()
        self.p_values_pileup.track_size = self.graph.node_indexes[-1]
        self.sample_pileup = None
        self.control_pileup = None

    def get_p_to_q_values_mapping(self):
        assert self.p_values_pileup is not None
        finder = PToQValuesMapper.from_p_values_pileup(
            self.p_values_pileup)
        self.p_to_q_values_mapping = finder.get_p_to_q_values()

    def get_q_values(self):
        assert self.p_values_pileup is not None
        assert self.p_to_q_values_mapping is not None
        finder = QValuesFinder(self.p_values_pileup,
                               self.p_to_q_values_mapping)
        self.q_values_pileup = finder.get_q_values()
        self.q_values_pileup.track_size = self.p_values_pileup.track_size
        self._reporter.add("qvalues", self.q_values_pileup)

    def call_peaks_from_q_values(self, experiment_info, config=None):
        assert self.q_values_pileup is not None
        caller = CallPeaksFromQvalues(
            self.graph, self.q_values_pileup,
            experiment_info, self.out_file_base_name,
            touched_nodes=self.touched_nodes,
            config=config
            )
        caller.callpeaks()
        self.max_path_peaks = caller.max_paths

    def save_max_path_sequences_to_fasta_file(self, file_name,
                                              sequence_retriever):
        assert self.max_path_peaks is not None
        f = open(self.out_file_base_name + file_name, "w")
        i = 0
        for max_path in self.max_path_peaks:
            seq = sequence_retriever.get_interval_sequence(max_path)
            f.write(">peak" + str(i) + " " +
                    max_path.to_file_line() + "\n" + seq + "\n")
            i += 1
        f.close()
        logging.info("Wrote max path sequences to fasta file: %s" % (
            self.out_file_base_name + file_name))

    @classmethod
    def run_from_intervals(
            cls, graph, sample_intervals,
            control_intervals=None, experiment_info=None,
            out_file_base_name="", has_control=True,
            linear_map=None, configuration=None, stop_after_p_values=False):
        caller = cls(graph, out_file_base_name)
        caller.sample_intervals = sample_intervals
        caller.control_intervals = control_intervals

        caller.run_pre_callpeaks(has_control, experiment_info,
                                 linear_map, configuration)
        caller.get_p_values()
        caller.p_values_pileup.track_size = graph.node_indexes[-1]
        if stop_after_p_values:
            np.save(out_file_base_name + "touched_nodes.npy",
                    np.array(list(caller.touched_nodes), dtype="int"))
            return caller.p_values_pileup.to_sparse_files(
                out_file_base_name + "pvalues")

        caller.get_p_to_q_values_mapping()
        caller.get_q_values()
        caller.call_peaks_from_q_values(experiment_info, configuration)
        return caller


class CallPeaksFromQvalues:
    def __init__(self, graph, q_values_pileup,
                 experiment_info,
                 out_file_base_name="",
                 cutoff=0.1, raw_pileup=None, touched_nodes=None,
                 config=None, q_values_max_path=False):
        self.graph = graph
        self.q_values = q_values_pileup
        self.info = experiment_info
        self.out_file_base_name = out_file_base_name
        self.cutoff = cutoff
        self.raw_pileup = raw_pileup
        self.touched_nodes = touched_nodes
        self.save_tmp_results_to_file = False
        self.graph_is_partially_ordered = False
        self.q_values_max_path = q_values_max_path

        if config is not None:
            self.cutoff = config.p_val_cutoff
            self.graph_is_partially_ordered = config.graph_is_partially_ordered
            self.save_tmp_results_to_file = config.save_tmp_results_to_file
        self._reporter = Reporter(self.out_file_base_name)

        self.info.to_file(self.out_file_base_name + "experiment_info.pickle")
        logging.info("Using p value cutoff %.4f" % self.cutoff)

    def __threshold(self):
        threshold = -np.log10(self.cutoff)
        logging.info("Thresholding peaks on q value %.4f" % threshold)
        self.pre_processed_peaks = self.q_values.threshold_copy(threshold)
        self._reporter.add("thresholded", self.pre_processed_peaks)

    def __postprocess(self):
        logging.info("Filling small Holes")
        if isinstance(self.pre_processed_peaks, DensePileup):
            self.pre_processed_peaks = SparseValues.from_dense_pileup(
                self.pre_processed_peaks.data._values)
        self.pre_processed_peaks = HolesCleaner(
            self.graph,
            self.pre_processed_peaks,
            self.info.read_length,
            self.touched_nodes
        ).run()
        self._reporter.add("hole_cleaned", self.pre_processed_peaks)

        logging.info("Not removing small peaks")
        self.filtered_peaks = self.pre_processed_peaks

    def __get_max_paths(self):
        logging.info("Getting maxpaths")
        if not self.q_values_max_path:
            _pileup = SparseValues.from_sparse_files(
                self.out_file_base_name+"direct_pileup")
            self.raw_pileup = _pileup
        else:
            _pileup = SparseValues.from_dense_pileup(
                self.q_values.data._values)
        logging.info("Running Sparse Max Paths")
        max_paths, sub_graphs = SparseMaxPaths(
            self.filtered_peaks, self.graph, _pileup).run()

        self._reporter.add("all_max_paths", max_paths)
        logging.info("All max paths found")

        # Create dense q
        if not isinstance(self.q_values, DensePileup):
            q_values = DensePileup(self.graph)
            q_values.data._values = self.q_values.to_dense_pileup(
                self.graph.node_indexes[-1])
            self.q_values = q_values

        for max_path in max_paths:
            max_path.set_score(np.max(
                self.q_values.data.get_interval_values(max_path)))
        pairs = list(zip(max_paths, sub_graphs))
        pairs.sort(key=lambda p: p[0].score, reverse=True)
        logging.info("N unfiltered peaks: %s", len(max_paths))
        pairs = [p for p in pairs if
                 p[0].length() >= self.info.fragment_length]
        logging.info("N filtered peaks: %s", len(pairs))
        self._reporter.add("sub_graphs", [pair[1] for pair in pairs])
        self.max_paths = [p[0] for p in pairs]
        self._reporter.add("max_paths", self.max_paths)

    def callpeaks(self):
        logging.info("Calling peaks")
        self.__threshold()
        self.__postprocess()
        self.__get_max_paths()

    def save_max_path_sequences_to_fasta_file(self, file_name, sequence_retriever):
        assert self.max_paths is not None, \
                "Max paths has not been found. Run peak calling first."
        assert sequence_retriever is not None
        # assert isinstance(sequence_retriever, vg.sequences.SequenceRetriever)
        f = open(self.out_file_base_name + file_name, "w")
        i = 0
        for max_path in self.max_paths:
            seq = sequence_retriever.get_interval_sequence(max_path)
            f.write(">peak" + str(i) + " " +
                    max_path.to_file_line() + "\n" + seq + "\n")
            i += 1
        f.close()
        logging.info("Wrote max path sequences to fasta file: %s" % (self.out_file_base_name + file_name))

    @staticmethod
    def intervals_to_fasta_file(interval_collection, out_fasta_file_name, sequence_retriever):
        f = open(out_fasta_file_name, "w")
        i = 0
        for max_path in interval_collection.intervals:
            seq = sequence_retriever.get_interval_sequence(max_path)
            f.write(">peak" + str(i) + " " +
                    max_path.to_file_line() + "\n" + seq + "\n")
            i += 1
            if i % 100 == 0:
                logging.info("Writing sequence # %d" % i)
        f.close()




import json
import offsetbasedgraph as obg
import logging
from collections import defaultdict
import numpy as np
from .nongraphpeaks import NonGraphPeakCollection, NonGraphPeak
from offsetbasedgraph import DirectedInterval


class Peak(obg.DirectedInterval):
    def __init__(self, start_position, end_position,
                 region_paths=None, graph=None, direction=None, score=0, unique_id=None):
        super().__init__(start_position, end_position,
                         region_paths, graph, direction)
        self.score = score
        self.unique_id = unique_id
        self.info = (0, 0)

    def __str__(self):
        return super().__str__() + " [%s]" % self.score

    @classmethod
    def from_interval_and_score(cls, interval, score):
        return cls(interval.start_position, interval.end_position,
                   interval.region_paths, interval.graph, interval.direction,
                   score)

    def set_score(self, score):
        assert isinstance(score, float), "Score %s is invalid, type %s" % (score, type(score))
        self.score = score

    def to_file_line(self):
        object = {"start": int(self.start_position.offset),
                  "end": int(self.end_position.offset),
                  "region_paths": [int(r) for r in self.region_paths],
                  "direction": int(self.direction),
                  "average_q_value": float(self.score),
                  "unique_id": str(self.unique_id),
                  "is_ambigous": int(self.info[1]),
                  "is_diff": int(self.info[0])
                  }
        try:
            d = json.dumps(object)
        except:
            for k, v in object.items():
                print(k, v, type(v))
            raise
        return d

    @classmethod
    def from_file_line(cls, line, graph=None):
        object = json.loads(line)
        unique_id = None
        if "unique_id" in object:
            unique_id = object["unique_id"]
        obj = cls(object["start"], object["end"], object["region_paths"],
                  direction=object["direction"], graph=graph,
                  score=object["average_q_value"],
                  unique_id=unique_id)
        obj.info = (object["is_diff"], object["is_ambigous"])
        return obj

    def to_approx_linear_peak(self, linear_path, chromosome):
        # Assuming graph is partially ordered DAG
        intersecting_nodes = set(self.region_paths).intersection(linear_path.nodes_in_interval())
        intersecting_nodes = sorted(list(intersecting_nodes))

        first_node = intersecting_nodes[0]
        start_offset = 0
        if first_node == self.region_paths[0]:
            start_offset = self.start_position.offset

        last_node = intersecting_nodes[-1]
        end_offset = 0
        if last_node == self.region_paths[-1]:
            end_offset = self.end_position.offset

        linear_start_pos = linear_path.get_offset_at_node(first_node) + start_offset
        linear_end_pos = linear_path.get_offset_at_node(last_node) + end_offset
        return NonGraphPeak(chromosome, linear_start_pos, linear_end_pos)


class PeakCollection(obg.IntervalCollection):
    interval_class = Peak

    @classmethod
    def _is_in_graph(cls, peak, chrom, start_offset, end_offset):
        if peak.chromosome != chrom:
            return False
        if (peak.start < start_offset or peak.end > end_offset):
            return False
        return True

    def create_node_index(self):
        # Creates an index from region path to intervals touching the rp,
        # making it fast to do approx. overlap
        index = defaultdict(list)
        for peak in self.intervals:
            for rp in peak.region_paths:
                index[abs(rp)].append(peak)

        self._index = index

    def approx_contains_part_of_interval(self, interval):
        assert hasattr(self, "_index"), "Create index first by calling create_node_index()"
        for rp in interval.region_paths:
            if len(self._index[abs(rp)]) > 0:
                return True

        return False

    def which_approx_contains_part_of_interval(self, interval):
        assert hasattr(self, "_index"), "Create index first by calling create_node_index()"
        for rp in interval.region_paths:
            containing = self._index[abs(rp)]
            if len(containing) > 0:
                return containing[0]

    @classmethod
    def create_from_nongraph_peak_collection(
            cls, ob_graph, peak_collection,
            linear_path_interval,
            graph_region=None):
        peaks = peak_collection.peaks
        intervals_on_graph = []
        i = 0
        graph_start_offset = 0 if graph_region is None else graph_region.start
        for peak in peaks:
            start = peak.start - graph_start_offset
            end = peak.end - graph_start_offset
            end = min(end, linear_path_interval.length())
            if graph_region is not None:
                if not cls._is_in_graph(peak, graph_region.chromosome,
                                        graph_region.start,
                                        graph_region.end):
                    logging.info("Filtered out peak")
                    continue
            if i % 100 == 0:
                logging.info("%d peaks processed" % (i))
            i += 1
            linear_interval = linear_path_interval.get_exact_subinterval(start, end)
            linear_interval.graph = ob_graph
            graph_peak = Peak.from_interval_and_score(
                linear_interval, peak.score)
            graph_peak.unique_id = peak.unique_id
            graph_peak.sequence = peak.sequence
            intervals_on_graph.append(graph_peak)

        return cls(intervals_on_graph)

    def contains_interval(self, interval):
        for i in self.intervals:
            if i == interval:
                return True

        return False

    def get_similar_intervals(self, interval, allowed_mismatches):
        similar = []
        for i in self.intervals:
            if i.is_approx_equal(interval, allowed_mismatches):
                similar.append(i)

        return similar

    def get_identical_intervals(self, other_peak_collection):
        identical_intervals = []
        for interval in self.intervals:
            if other_peak_collection.contains_interval(interval):
                identical_intervals.append(interval)

        return identical_intervals

    def get_peaks_not_in_other_collection(self, other_collection,
                                          allowed_mismatches=0):
        out = []
        for peak in self:
            similar = other_collection.get_similar_intervals(
                peak, allowed_mismatches)
            if len(similar) == 0:
                out.append(peak)

        return out

    def get_overlapping_intervals(self, interval, minimum_overlap=1):
        overlapping = []
        for i in self.intervals:
            if i.overlaps(interval, minimum_overlap=minimum_overlap):
                overlapping.append(i)
        return overlapping

    def to_approx_linear_peaks(self, linear_path, chromosome):
        linear_peaks = []
        i = 0
        for peak in self.intervals:
            if i % 500 == 0:
                logging.info("Converting peak %d" % i)
            i += 1
            linear_peaks.append(peak.to_approx_linear_peak(linear_path, chromosome))

        return NonGraphPeakCollection(linear_peaks)


    @classmethod
    def from_fasta_file(cls, file_name, graph=None):
        f = open(file_name)
        peaks = []
        while True:
            header = f.readline()
            sequence = f.readline()
            if not sequence:
                break

            header = header.split(maxsplit=1)
            id = header[0].replace(">", "")
            interval_json = header[1]
            peak = Peak.from_file_line(interval_json)
            peak.unique_id = id
            peak.sequence = sequence
            peak.graph = graph
            peaks.append(peak)
        avg_peak_size = np.mean([p.length() for p in peaks])
        logging.info("Avg peak size: %2.f" % avg_peak_size)

        return cls(peaks)




import numpy as np
import logging
import offsetbasedgraph as obg
from pyvg.conversion import vg_json_file_to_interval_collection
from . import ExperimentInfo, CallPeaks, Configuration
from .sparsepvalues import PToQValuesMapper
from .sparsediffs import SparseValues

logging.basicConfig(level=logging.INFO,
                    format="%(asctime)s, %(levelname)s: %(message)s")


class MultipleGraphsCallpeaks:

    def __init__(self, graph_names, graph_file_names,
                 samples,
                 controls, linear_maps,
                 fragment_length, read_length,
                 has_control=False,
                 sequence_retrievers=None,
                 out_base_name="multigraphs_",
                 skip_filter_duplicates=False,
                 save_tmp_results_to_file=False,
                 stop_after_p_values=False,
                 min_background=None
                 ):

        self._config = Configuration(
            skip_read_validation=True, save_tmp_results_to_file=save_tmp_results_to_file,
            skip_filter_duplicates=skip_filter_duplicates, p_val_cutoff=0.05,
            graph_is_partially_ordered=True,
            use_global_min_value=min_background)
        self.names = graph_names
        self.graph_file_names = graph_file_names
        self.fragment_length = fragment_length
        self.read_length = read_length
        self.linear_maps = linear_maps
        self.has_control = has_control
        self._base_name = out_base_name
        self.sequence_retrievers = sequence_retrievers
        self.samples = samples
        self.controls = controls
        self.stop_after_p_values = stop_after_p_values
        if self.stop_after_p_values:
            logging.info("Will only run until p-values have been computed.")
        #self.run()

    @classmethod
    def count_number_of_unique_reads(cls, sample_reads):
        n_unique = 0
        for reads in sample_reads:
            logging.info("Processing sample")

            interval_hashes = set()
            n_duplicates = 0
            i = 0
            for interval in reads.intervals:
                if i % 20000 == 0:
                    logging.info("%d reads processed" % i)
                i += 1
                hash = interval.hash()
                if hash in interval_hashes:
                    n_duplicates += 1
                else:
                    n_unique += 1
                interval_hashes.add(hash)

            print("Found %d duplicates" % n_duplicates)

        logging.info("In total %d unique reads" % n_unique)
        return n_unique

    @classmethod
    def from_file_base_names(cls):
        # Create lists of names, graphs, sample, control linear maps
        pass

    def run(self):
        self.run_to_p_values()
        if self.stop_after_p_values:
            logging.info("Stopping, as planned, after p-values")
            return
        self.create_joined_q_value_mapping()
        self.run_from_p_values()

    def run_to_p_values(self):
        for name, graph_file_name, sample, control, lin_map in \
            zip(self.names, self.graph_file_names,
                self.samples, self.controls, self.linear_maps):
            logging.info("Running %s" % name)
            ob_graph = obg.GraphWithReversals.from_unknown_file_format(
                graph_file_name)
            if isinstance(sample, obg.IntervalCollection):
                logging.info("Sample is already intervalcollection.")
            elif sample.endswith(".intervalcollection"):
                sample = obg.IntervalCollection.create_generator_from_file(sample, graph=ob_graph)
                control = obg.IntervalCollection.create_generator_from_file(control, graph=ob_graph)
            else:
                logging.info("Creating interval collections from files")
                sample = vg_json_file_to_interval_collection(sample, ob_graph)
                control = vg_json_file_to_interval_collection(control, ob_graph)

            graph_size = ob_graph.number_of_basepairs()
            info = ExperimentInfo(
                graph_size, self.fragment_length, self.read_length)
            CallPeaks.run_from_intervals(
                ob_graph, sample, control, info, self._base_name + name + "_",
                self.has_control,
                lin_map, self._config,
                stop_after_p_values=True
            )
            logging.info("Done until p values.")

    def create_joined_q_value_mapping(self):
        mapper = PToQValuesMapper.from_files(self._base_name)
        self._q_value_mapping = mapper.get_p_to_q_values()

    def run_from_p_values(self, only_chromosome=None):
        for i, name in enumerate(self.names):
            if only_chromosome is not None:
                if only_chromosome != name:
                    logging.info("Skipping %s" % str(name))
                    continue
            graph_file_name = self.graph_file_names[i]
            ob_graph = obg.GraphWithReversals.from_unknown_file_format(
                graph_file_name)
            graph_size = sum(
                block.length() for block in ob_graph.blocks.values())
            info = ExperimentInfo(
                graph_size, self.fragment_length, self.read_length)
            assert ob_graph is not None
            caller = CallPeaks(ob_graph, self._base_name + name + "_")
            caller.p_to_q_values_mapping = self._q_value_mapping
            caller.p_values_pileup = SparseValues.from_sparse_files(
                self._base_name + name + "_" + "pvalues")
            caller.touched_nodes = set(np.load(
                self._base_name + name + "_" + "touched_nodes.npy"))
            caller.get_q_values()
            caller.call_peaks_from_q_values(
                experiment_info=info, config=self._config)
            if self.sequence_retrievers is not None:
                caller.save_max_path_sequences_to_fasta_file(
                    "sequences.fasta", self.sequence_retrievers.__next__())



import pickle
import logging
from itertools import chain
import numpy as np
from scipy.stats import poisson
from collections import defaultdict
from offsetbasedgraph import Interval, IntervalCollection


class ValuedIndexes(object):
    def __init__(self, indexes, values, start_value, length):

        if isinstance(indexes, list):
            indexes = np.array(indexes, dtype="int")

        if isinstance(values, list):
            values = np.array(values)

        assert type(indexes) == np.ndarray
        assert type(values) == np.ndarray
        # assert indexes.size == values.size
        self.length = length
        self.__values = None
        self.__indexes = None
        self.__start_value = 0
        self._all_values = np.zeros(0)
        self._all_indexes = np.zeros(0)

        self.values = values
        self.indexes = indexes
        self.start_value = start_value

        self._all_values[0] = start_value

        self.__tmp_starts = []
        self.__tmp_values = []
        self.__tmp_end = 0

    @property
    def indexes(self):
        #assert np.all(self.__indexes == self._all_indexes[1:-1])
        return self.__indexes

    @indexes.setter
    def indexes(self, new_indexes):
        self.__indexes = new_indexes
        if len(new_indexes) != len(self._all_indexes) - 2:
            self._all_indexes = np.zeros(len(new_indexes)+2, dtype=np.int)
        self._all_indexes[1:-1] = new_indexes
        self._all_indexes[-1] = self.length

    @property
    def values(self):
        return self.__values

    @values.setter
    def values(self, new_values):
        self.__values = new_values
        if len(new_values) != len(self._all_values) - 1 or new_values.ndim != self._all_values.ndim:
            if new_values.ndim == 2:
                self._all_values = np.zeros((len(new_values)+1, 2))
            else:
                self._all_values = np.zeros(len(new_values)+1)

            self._all_values[0] = self.start_value

        self._all_values[1:] = new_values

    @property
    def start_value(self):
        return self.__start_value

    @start_value.setter
    def start_value(self, value):
        self._all_values[0] = value
        self.__start_value = value

    def set_single_value(self, index, value):
        self.__values[index] = value
        self._all_values[index+1] = value

    def all_values(self):
        return self._all_values
        #return np.insert(self.values, 0, self.start_value)

    def all_idxs(self):
        return self._all_indexes
        #return np.append(np.insert(self.indexes, 0, 0), self.length)

    def __eq__(self, other):
        if not np.allclose(self.values, other.values):
            return False

        if not np.allclose(self.indexes, other.indexes):
            return False

        if isinstance(self.start_value, np.ndarray):
            if not np.allclose(self.start_value, other.start_value):
                return False
        else:
            if not np.isclose(self.start_value, other.start_value):
                return False
        return self.length == other.length

    def __str__(self):
        return "VI(indices: %s, values: %s, start: %s, length: %s)" % (
            self.indexes, self.values,
            self.start_value, self.length)

    __repr__ = __str__

    def sum(self):
        lengths = np.diff(self.all_idxs())
        return np.sum(lengths*self.all_values())

    def max_value(self):
        return np.max(self.all_values())

    def mean(self):
        return self.sum()/self.length

    def get_subset(self, start, end):
        assert start >= 0
        assert end <= self.length
        if not self.indexes.size:
            return ValuedIndexes(np.array([], dtype="int"),
                                 np.array([]), self.start_value,
                                 end-start)
        length = end-start
        indexes = self.indexes-start
        above_mask = indexes > 0
        above = np.nonzero(above_mask)[0]

        if not above.size:
            start_value = self.values[-1] if self.values.size else self.start_value
        elif above[0] == 0:
            start_value = self.start_value
        else:
            start_value = self.values[above[0]-1]
        if not above.size:
            return self.__class__(np.array([], dtype="int"),
                                  np.array([]),
                                  start_value,
                                  length)
        inside_mask = np.logical_and(above_mask,
                                     indexes < length)
        subset_indexes = indexes[inside_mask]
        subset_values = self.values[inside_mask]
        return self.__class__(subset_indexes,
                              subset_values, start_value,
                              length)

    def scale(self, scale):
        self.values = self.values*scale
        self.start_value = self.start_value*scale

    def tmp_set_interval_value(self, start, end, value):
        self.__tmp_starts.append(start)
        self.__tmp_values.append(value)
        self.__tmp_end = max(self.__tmp_end, end)

    def fix_tmp_values(self):
        assert self.__tmp_starts[0] == 0
        self.start_value = self.__tmp_values[0]
        self.length = self.__tmp_end
        if len(self.__tmp_starts) < 2:
            self.values = np.array([])
            self.indexes = np.array([], dtype="int")
        else:
            self.indexes = np.array(self.__tmp_starts[1:])
            self.values = np.array(self.__tmp_values[1:])
        self.__tmp_values = []
        self.__tmp_starts = []
        self.__tmp_end = 0

    def set_interval_value(self, start, end, value):
        # ONly used for filling holes. TODO rename
        if start == 0:
            self.start_value = value
        else:
            idx = np.nonzero(self.indexes == start)
            assert len(idx) == 1
            idx = idx[0]
            #self.values[idx] = value
            self.set_single_value(idx, value)

            assert end == self.length or np.any(self.indexes[idx[0]+1] == end)

    def set_interval_value_on_right_empty_area(self, start, end, value):
        # Requires that everything after the end is is empty
        if start == 0:
            self.start_value = value
        else:

            idx = np.where(start == self.indexes)[0]
            assert len(idx) <= 1
            if len(idx) == 1:
                self.values[idx] = value
            else:
                self.indexes = np.append(self.indexes, [start])
                self.values = np.append(self.values, [value])

        if end < self.length:
            self.indexes = np.append(self.indexes, [end])
            self.values = np.append(self.values, [0])

    def threshold_copy(self, cutoff):
        new_vi = self.__class__(np.copy(self.indexes),
                                self.values >= cutoff,
                                self.start_value >= cutoff,
                                self.length)
        new_vi.sanitize()
        return new_vi

    def threshold(self, cutoff):
        self.values = self.values >= cutoff
        self.start_value = self.start_value >= cutoff

    def sanitize(self):
        #diffs = np.diff(np.insert(self.values, 0, self.start_value))
        #assert np.all(self.all_values() == np.insert(self.values, 0, self.start_value))
        all_vals = self.all_values()

        diffs = np.diff(all_vals)
        changes = np.where(diffs != 0)[0]
        if len(self.values) == 0:
            return

        self.values = self.values[changes]
        self.indexes = self.indexes[changes]

    def sanitize_indices(self):
        indexes = self.all_idxs()
        values = self.all_values()
        diffs = np.where(np.diff(indexes) > 0)[0]

        self.indexes = indexes[diffs[1:]]
        self.values = values[diffs[1:]]
        self.start_value = values[diffs[0]]

    def find_valued_areas(self, value):
        all_indexes = self.all_idxs()
        values = self.all_values()
        idxs = np.where(values == value)[0]
        starts = all_indexes[idxs]
        ends = all_indexes[idxs+1]
        areas = list(chain(*zip(starts, ends)))
        return areas

    @classmethod
    def maximum(cls, vi_a, vi_b):
        a = vi_a.all_idxs()[:-1]*2
        b = vi_b.all_idxs()[:-1]*2+1
        all_idxs = np.concatenate([a, b])
        all_idxs.sort()
        vi_list = [vi_a, vi_b]
        values_list = []
        for i, vi in enumerate(vi_list):
            idxs = np.nonzero((all_idxs % 2) == i)[0]
            all_values = vi.all_values()
            value_diffs = np.diff(all_values)
            values = np.zeros(all_idxs.shape)
            values[idxs[1:]] = value_diffs
            values[idxs[0]] = all_values[0]
            values_list.append(values.cumsum())
        values = np.maximum(values_list[0], values_list[1])
        idxs = all_idxs // 2
        empty_ends = np.nonzero(np.diff(idxs) == 0)[0]
        max_values = np.maximum(values[empty_ends], values[empty_ends+1])
        values[empty_ends+1] = max_values
        values[empty_ends] = max_values
        obj = cls(idxs[1:], values[1:], values[0], vi_a.length)
        obj.sanitize()
        return obj

    @classmethod
    def combine(cls, vi_a, vi_b):
        a = vi_a.all_idxs()[:-1]*2
        b = vi_b.all_idxs()[:-1]*2+1
        all_idxs = np.concatenate([a, b])
        all_idxs.sort()
        vi_list = [vi_a, vi_b]
        values_list = []
        for i, vi in enumerate(vi_list):
            idxs = np.nonzero((all_idxs % 2) == i)[0]
            all_values = vi.all_values()
            value_diffs = np.diff(all_values)
            values = np.zeros(all_idxs.shape)
            values[idxs[1:]] = value_diffs
            values[0] = all_values[0]
            values_list.append(values.cumsum())

        values = np.array([values_list[0], values_list[1]])
        idxs = all_idxs // 2
        unique_idxs = np.append(np.nonzero(np.diff(idxs))[0], len(idxs)-1)
        idxs = idxs[unique_idxs]
        values = values[:, unique_idxs]
        obj = cls(idxs[1:], np.transpose(values[:, 1:]),
                  values[:, 0], vi_a.length)
        return obj

    def trunctate(self, min_value):
        self.values = np.maximum(self.values, min_value)
        self.start_value = max(min_value, self.start_value)
        self.sanitize()

    @classmethod
    def empty(cls, length):
        return cls(np.array([], dtype="int"),
                   np.array([]),
                   0,
                   length)

    def __iter__(self):
        return zip(
            chain([0], self.indexes),
            chain(self.indexes, [self.length]),
            chain([self.start_value], self.values)
            )


class BinaryIndexes(object):
    def __init__(self, starts, ends, length):
        self.starts = starts
        self.ends = ends
        self.length = length

    def add_interval(self, start, end):
        self.starts.append(start)
        self.ends.append(end)
        self.starts.sort


class SparsePileupData(dict):

    def __init__(self, *args, **kwargs):
        self.graph = kwargs["graph"]
        self.min_value = 0
        del kwargs["graph"]
        super(SparsePileupData, self).__init__(*args, **kwargs)
        self.default_value = 0

    def __getitem__(self, item):
        if item not in self:
            value = ValuedIndexes.empty(self.graph.node_size(item))
            value.start_value = max(self.min_value, self.default_value)
            self.__setitem__(item, value)

        return super(SparsePileupData, self).__getitem__(item)

    def __setitem__(self, key, value):
        super(SparsePileupData, self).__setitem__(key, value)

    def trunctate(self, min_value):
        for key in self.keys():
            self.__getitem__(key).trunctate()
        self.min_value = max(self.min_value, min_value)

    def values(self):
        return (self.__getitem__(key) for key in self.graph.blocks)

    # def items(self):
    #    return ((key, self.__getitem__(key)) for key in self.graph.blocks)

    def all_items(self):
        # Also gives rps with zero
        return ((key, self.__getitem__(key)) for key in self.graph.blocks)


class NpSparseAreasDict(dict):
    def __init__(self, diffs, changes):
        self._diffs = diffs
        self._changes = changes

    def __getitem__(self, item):
        pass

class SparseAreasDict(dict):
    """
    Used as return object in SparsePileup.get_valued_areas() to speedup.
    Assumes value in empty nodes, i.e. nodes not in dict are completely filled
    """
    def __init__(self, *args, **kwargs):
        self.graph = kwargs["graph"]
        self.min_value = 0
        del kwargs["graph"]
        super(SparseAreasDict, self).__init__(*args, **kwargs)

    def __getitem__(self, item):
        if item not in self:
            value = [0, self.graph.node_size(item)]
            self.__setitem__(item, value)

        return super(SparseAreasDict, self).__getitem__(item)

    def values(self):
        raise NotImplementedError()

    def items(self):
        return ((key, self.__getitem__(key)) for key in self.graph.blocks)


class SparsePileup:
    def __init__(self, graph):
        logging.info("Initing sparsepileup")
        self.graph = graph
        self.data = SparsePileupData(graph=self.graph)

        #self.data = {rp: ValuedIndexes.empty(graph.node_size(rp))
        #             for rp in self.graph.blocks}
        logging.info("Sparsepileup inited")
        #self.graph.assert_correct_edge_dicts()

    def __eq__(self, other):
        for node_id, vi in other.data.all_items():
            if self.data[node_id] != vi:
                return False
        return True

    @classmethod
    def from_base_value(cls, graph, base_value):
        pileup = cls(graph)
        pileup.update_max_value(base_value)
        return pileup

    def sum(self):
        return np.sum([values.sum() for node, values in self.data.items()])

    def mean(self):
        graph_size = sum([self.graph.node_size(b) for b in self.graph.blocks])
        mean = self.sum() / graph_size
        return mean

    def scale(self, scale):
        [vi.scale(scale) for vi in self.data.values()]

    def fill_small_wholes(self, max_size, write_holes_to_file=None, touched_nodes=None):
        cleaner = HolesCleaner(self, max_size, touched_nodes=touched_nodes)
        areas = cleaner.run()
        n_filled = 0

        hole_intervals = []

        for node_id in areas.areas:
            if touched_nodes is not None:
                if node_id not in touched_nodes:
                    continue

            starts = areas.get_starts(node_id)
            ends = areas.get_ends(node_id)
            for start, end in zip(starts, ends):
                self.data[node_id].set_interval_value(start, end, True)
                logging.debug("Filling hole %s, %d, %d" % (
                    node_id, start, end))
                n_filled += 1
                assert end - start <= max_size
                hole_intervals.append(Interval(start, end, [node_id]))

        logging.info(
            "Filled %d small holes (splitted into holes per node)" % n_filled)

        if write_holes_to_file is not None:
            intervals = IntervalCollection(hole_intervals)
            intervals.to_file(write_holes_to_file, text_file=True)

        self.sanitize()

    def sanitize(self):
        for node_id, valued_indexes in self.data.items():
            valued_indexes.sanitize()
            assert node_id in self.graph.blocks

    def find_valued_areas(self, value, only_check_nodes=None):
        if value:
            return {node_id: self.data[node_id].find_valued_areas(value)
                    for node_id, valued_indexes in self.data.items()}
        else:

            if only_check_nodes is not None:
                nodes = only_check_nodes
            else:
                nodes = self.graph.blocks

            return SparseAreasDict({node_id: self.data[node_id].find_valued_areas(value)
                                   for node_id in self.data
                                    }, graph=self.graph)

            """
            return {node_id: (self.data[node_id].find_valued_areas(value)
                              if node_id in self.data
                              else [0, self.graph.node_size(node_id)])
                    for node_id in nodes}
            """

    def set_valued_intervals(self, node_id, valued_indexes):
        assert node_id in self.graph.blocks
        self.data[node_id] = valued_indexes

    @classmethod
    def from_areas_collection(cls, graph, areas_list):
        starts_dict = defaultdict(list)
        ends_dict = defaultdict(list)
        for areas in areas_list:
            for rp in areas.areas:
                starts_dict[rp].extend(areas.get_starts(rp))
                ends_dict[rp].extend(areas.get_ends(rp))
        starts_dict = {rp: np.array(v) for rp, v in starts_dict.items()}
        ends_dict = {rp: np.array(v) for rp, v in ends_dict.items()}
        return cls.from_starts_and_ends(graph, starts_dict,
                                        ends_dict, dtype=int)

    @classmethod
    def from_valued_areas(cls, graph, valued_areas, touched_nodes = None):
        pileup = cls(graph)
        i = 0

        if touched_nodes is None:
            nodes = graph.blocks
        else:
            nodes = touched_nodes

        for rp in nodes:
            if i % 100000 == 0:
                logging.info("Creating sparse from valued areas for node %d" % i)
            i += 1

            length = graph.blocks[rp].length()
            starts = valued_areas.get_starts_array(rp, node_size=length)
            if len(starts) == 0:
                continue

            ends = valued_areas.get_ends_array(rp, node_size=length)
            if len(starts) == 0 and len(ends) == 0:
                continue

            indexes, values = starts_and_ends_to_sparse_pileup(
                starts,
                ends)
            if not indexes.size:
                #assert not valued_areas.has_anything_on_node(rp)
                continue
            start_value = 0
            if indexes[0] == 0:
                start_value = values[0]
                indexes = indexes[1:]
                values = values[1:]

            if len(indexes) > 0:
                if indexes[-1] == length:
                    indexes = indexes[:-1]
                    values = values[:-1]

            pileup.data[rp] = ValuedIndexes(
                indexes, values, start_value, length)

        return pileup

    @classmethod
    def from_starts_and_ends(cls, graph, starts, ends, dtype=bool):
        pileup = cls(graph)
        for rp in starts:
            assert rp in graph.blocks
            indexes, values = starts_and_ends_to_sparse_pileup(
                starts[rp], ends[rp])
            start_value = dtype(False)
            length = graph.blocks[rp].length()
            if indexes[0] == 0:
                start_value = values[0]
                indexes = indexes[1:]
                values = values[1:]

            if len(indexes) > 0:
                if indexes[-1] == length:
                    indexes = indexes[:-1]
                    values = values[:-1]

            pileup.data[rp] = ValuedIndexes(
                indexes, values, start_value, length)

        logging.info("Number of elements in pileup: %d" % len(pileup.data))
        return pileup

    @classmethod
    def from_intervals(cls, graph, intervals):
        starts, ends = intervals_to_start_and_ends(graph, intervals)
        return cls.from_starts_and_ends(graph, starts, ends)

    def __str__(self):
        return "\n".join(
            "%s: %s, %s, %s" % (node_id, valued_indexes.indexes, valued_indexes.values, valued_indexes.start_value)
            for node_id, valued_indexes in self.data.items())

    __repr__ = __str__

    def threshold_copy(self, cutoff):
        new_data = {node_id: vi.threshold_copy(cutoff)
                    for node_id, vi in self.data.items()}

        pileup = self.__class__(self.graph)
        pileup.data = new_data
        return pileup

    def threshold(self, cutoff):
        n = 0
        logging.info("Thresholding. Number of nodes to threshold: %d" % len(self.data))
        for node, valued_indexes in self.data.items():
            if n % 25000 == 0:
                logging.info("Thresholding node %d" % n)
            n += 1
            valued_indexes.threshold(cutoff)
            valued_indexes.sanitize()

    def add_intervals(self, intervals):
        [self.add_interval(interval)
         for interval in intervals]

    def add_interval(self, interval):
        assert all(region_path in self.graph.blocks for
                   region_path in interval.region_paths)
        for i, region_path in enumerate(interval.region_paths):
            start = 0
            end = self.graph.node_size(region_path)
            if i == 0:
                start = interval.start_position.offset
            if i == len(interval.region_paths)-1:
                end = interval.end_position.offset
            self.data[region_path].add_interval(
                start, end)

    def add_area(self, region_path, start, end, value):
        """
        NAIVE method to add single area to valued_inexes
        Assumes it comes from a complete set of areas
        """
        vi = self.data[region_path]
        vi.tmp_set_interval_value(start, end, value)

    def fix_tmp_values(self):
        logging.info("Fixing tmp values")
        for vi in self.data.values():
            vi.fix_tmp_values()

    @classmethod
    def from_bed_graph(cls, graph, filename):
        pileup = super().from_bed_graph(graph, filename)
        assert isinstance(pileup, cls)
        pileup.fix_tmp_values()
        return pileup

    def add_areas(self, areas):
        for area, intervals in areas.items():
            self.data[area].add_interval(
                intervals[0], intervals[1])

    def set_areas_value(self, areas, value):
        for area, intervals in areas.items():
            for i in range(len(intervals)//2):
                self.data[area].set_interval_value(
                    intervals[i], intervals[i+1], value)

    def set_interval_value(self, interval, value):
        assert all(region_path in self.graph.blocks for
                   region_path in interval.region_paths)
        for i, region_path in enumerate(interval.region_paths):
            start = 0
            end = self.graph.node_size(region_path)
            if i == 0:
                start = interval.start_position.offset
            if i == len(interval.region_paths)-1:
                end = interval.end_position.offset
            self.data[region_path].set_interval_value(
                start, end, value)

    @classmethod
    def from_bed_file(cls, graph, filename):
        f = open(filename, "r")
        starts = defaultdict(list)
        ends = defaultdict(list)
        for line in f:
            if line.startswith("track"):
                continue

            data = line.split()
            block_id = int(data[0])
            starts[block_id].append(int(data[1]))
            ends[block_id].append(int(data[2]))
        starts = {block_id: np.array(start_list) for block_id, start_list in starts.items() if start_list}
        ends = {block_id: np.array(ends_list) for block_id, ends_list in ends.items() if ends_list}
        return cls.from_starts_and_ends(graph, starts, ends)

    def to_bed_file(self, filename):
        f = open(filename, "w")
        areas = self.find_valued_areas(True)
        for node_id, idxs in areas.items():
            for i in range(len(idxs)//2):
                interval = (node_id, idxs[2*i], idxs[2*i+1])
                f.write("%s\t%s\t%s\t.\t.\t.\n" % interval)
        f.close()
        return filename

    def set_to_false(self):
        for valued_indexes in self.data.values():
            valued_indexes.start_value = False
            valued_indexes.indexes = np.array([], dtype="int")
            valued_indexes.values = np.array([], dtype="bool")

    def remove_small_peaks(self, min_size):
        """
        areas = self.find_valued_areas(True)
        intervals = self.areas_to_intervals(areas, include_partial_stubs=False)
        large_intervals = [interval for interval in intervals
                           if interval.length() >= min_size]
        for i in [interval for interval in intervals
                  if interval.length() < min_size]:
            print(i)
        return self.from_intervals(self.graph, large_intervals)
        """
        logging.info("Initing cleaner")
        cleaner = PeaksCleaner(self, min_size)
        logging.info("Running cleaner")
        areas = cleaner.run()
        logging.info("Removing emty areas")
        #for node, startends in areas.areas.items():
        #    if len(startends) == 0:
        #        del areas[node]

        logging.info("Creating pileup using results from cleaner")
        pileup = self.from_areas_collection(self.graph, [areas])
        logging.info("Tresholding")
        pileup.threshold(0.5)
        return pileup

    def update_max(self, other):
        for key, valued_indexes in self.data.items():
            self.data[key] = ValuedIndexes.maximum(
                valued_indexes, other.data[key])

    def update_max_value(self, min_value):

        if isinstance(self.data, SparsePileupData):
            self.data.trunctate(min_value)
        else:
            for valued_indexes in self.data.values():
                valued_indexes.trunctate(min_value)

    def to_bed_graph(self, filename):
        f = open(filename, "w")
        for node_id, valued_indexes in self.data.items():
            for t in valued_indexes:
                f.write("%s\t%s\t%s\t%s\n" % ((node_id,) + t))
        f.close()
        self.filename = filename
        self.is_written = True
        return filename

    def set_sorted_interval_values(self, intervals, values):
        # Requires intervals to be sorted within nodes
        for j, interval in enumerate(intervals):
            logging.info("Setting sorted interval %d/%d" % (j, len(intervals)))
            for i, rp in enumerate(interval.region_paths):
                start = 0
                if i == 0:
                    start = interval.start_position.offset
                end = self.graph.node_size(rp)
                if i + 1 == len(interval.region_paths):
                    end = interval.end_position.offset
                self.data[rp].set_interval_value_on_right_empty_area(start, end, values[j])

    @classmethod
    def from_pickle(cls, file_name, graph):
        with open("%s" % file_name, "rb") as f:
            data = pickle.loads(f.read())
            #assert isinstance(SparsePileupData, cls)
            obj = cls(graph)
            obj.data = data
            return obj

    def to_pickle(self, file_name):
        with open("%s" % file_name, "wb") as f:
            pickle.dump(self.data, f)


class SparseControlSample(SparsePileup):
    def get_p_dict(self):
        p_value_dict = defaultdict(dict)
        count_dict = defaultdict(int)
        baseEtoTen = np.log(10)
        for node_id, valued_indexes in self.data.items():
            start = 0
            val = valued_indexes.start_value
            for start, end, val in valued_indexes:
                if val[1] not in p_value_dict[val[0]]:
                    log_p_val = poisson.logsf(val[1], val[0])
                    p_value_dict[val[0]][val[1]] = -log_p_val/baseEtoTen
                p = p_value_dict[val[0]][val[1]]
                count_dict[p] += end-start

        self.p_value_dict = p_value_dict
        self.count_dict = count_dict

    def get_p_to_q_values(self):
        p_value_counts = self.count_dict
        p_to_q_values = {}
        sorted_p_values = sorted(p_value_counts.keys(), reverse=True)
        rank = 1
        # logN = np.log10(self.graph.get_size())
        logN = np.log10(sum(p_value_counts.values()))
        pre_q = None
        for p_value in sorted_p_values:
            value_count = p_value_counts[p_value]
            q_value = p_value + (np.log10(rank) - logN)
            if np.isclose(p_value, 2.1326711212014025):
                print(q_value, logN, rank, np.log10(rank))
            if rank == 1:
                q_value = max(0.0, q_value)
            else:
                q_value = max(0.0, min(pre_q, q_value))
            p_to_q_values[p_value] = q_value
            pre_q = q_value
            rank += value_count
        self.p_to_q_values = p_to_q_values

    def get_q_values(self):
        def translation(x):
            return self.p_to_q_values[
                self.p_value_dict[x[0]][x[1]]]
        # f = np.vectorize(translation)
        for valued_indexes in self.data.values():

            valued_indexes.start_value = translation(
                valued_indexes.start_value)

            if valued_indexes.values.size:
                new_values = np.apply_along_axis(
                    translation, 1, valued_indexes.values)
                valued_indexes.values = new_values

            valued_indexes.sanitize()

    def get_scores(self):
        self.get_p_dict()
        self.get_p_to_q_values()
        self.get_q_values()

    @classmethod
    def from_sparse_control_and_sample(cls, control, sample):
        logging.info("Creating pileup by combining control and sample")
        logging.info("N nodes in control: %d" % len(control.data))
        sparse_pileup = cls(sample.graph)
        sparse_pileup.data = {
            node_id: ValuedIndexes.combine(
                control.data[node_id], sample.data[node_id])
            for node_id in control.data}
        return sparse_pileup

    @classmethod
    def from_control_and_sample(cls, control, sample):
        sparse_pileup = cls(control.graph)
        for node_id in sparse_pileup.graph.blocks:
            control_counts = control.get_count_array(node_id)
            sample_counts = sample.get_count_array(node_id)
            control_diffs = control_counts[1:]-control_counts[:-1]
            sample_diffs = sample_counts[1:]-sample_counts[:-1]
            changes = np.where(np.logical_or(control_diffs != 0,
                                             sample_diffs))[0] + 1
            vals = np.column_stack([control_counts[changes],
                                    np.floor(sample_counts[changes])])
            init_value = np.array([control_counts[0], sample_counts[0]])
            valued_indexes = ValuedIndexes(
                 changes, vals, init_value, control_counts.shape[0])
            assert not np.any(vals[:, 0] == 0)
            sparse_pileup.set_valued_intervals(node_id, valued_indexes)
        return sparse_pileup


def intervals_to_start_and_ends(graph, intervals):
    # Returns two dict on rp => positions (start/ends)
    starts = defaultdict(list)
    ends = defaultdict(list)

    for interval in intervals:
        for i, region_path in enumerate(interval.region_paths):
            start = 0
            end = graph.node_size(region_path)
            if i == 0:
                start = interval.start_position.offset
            if i == len(interval.region_paths)-1:
                end = interval.end_position.offset

            if region_path < 0:
                new_start = graph.node_size(region_path) - end
                new_end = graph.node_size(region_path) - start
                start = new_start
                end = new_end

                region_path = -region_path

            starts[region_path].append(start)
            ends[region_path].append(end)

    for rp in starts:
        starts[rp] = np.array(starts[rp])

    for rp in ends:
        ends[rp] = np.array(ends[rp])

    return starts, ends


def filter_pileup_duplicated_position(positions, values):
    equal_to_previous = (positions[:-1] == positions[1:])
    true_array = np.ones(len(positions))
    true_array[np.where(equal_to_previous)] = False
    return positions[np.where(true_array)], values[np.where(true_array)]



import logging

import offsetbasedgraph as obg
from pyvg.conversion import vg_json_file_to_interval_collection
from pyvg.sequences import SequenceRetriever

from . import ExperimentInfo, Configuration, CallPeaks
from .multiplegraphscallpeaks import MultipleGraphsCallpeaks


def run_callpeaks(ob_graph,
                  sample_file_name, control_file_name,
                  vg_graph_file_name,
                  out_name="real_data_",
                  has_control=True,
                  limit_to_chromosomes=False,
                  fragment_length=135, read_length=36,
                  linear_map_file_name=False):

    logging.info("Running from gam files")

    # Detect file format
    if isinstance(sample_file_name, obg.IntervalCollection):
        sample_intervals = sample_file_name
        control_intervals = control_file_name
    elif sample_file_name.endswith(".json"):
        sample_intervals = vg_json_file_to_interval_collection(
            sample_file_name, ob_graph)
        control_intervals = vg_json_file_to_interval_collection(
            control_file_name, ob_graph)
    else:
        try:
            sample_intervals = obg.IntervalCollection.from_file(
                sample_file_name, graph=ob_graph)
            control_intervals = obg.IntervalCollection.from_file(
                control_file_name, graph=ob_graph)
        except OSError:
            sample_intervals = obg.IntervalCollection.from_file(
                sample_file_name, graph=ob_graph, text_file=True)
            control_intervals = obg.IntervalCollection.from_file(
                control_file_name, graph=ob_graph, text_file=True)

    graph_size = ob_graph.number_of_basepairs()
    logging.info("Number of base pairs in graph: %d" % graph_size)

    experiment_info = ExperimentInfo(graph_size, fragment_length, read_length)
    config = Configuration(
        skip_read_validation=True, save_tmp_results_to_file=True,
        skip_filter_duplicates=False, p_val_cutoff=0.05,
        graph_is_partially_ordered=True)

    caller = CallPeaks.run_from_intervals(
        ob_graph, sample_intervals, control_intervals,
        experiment_info=experiment_info,
        out_file_base_name=out_name, has_control=has_control,
        linear_map=linear_map_file_name,
        configuration=config
    )
    if vg_graph_file_name != "None":
        try:
            retriever = SequenceRetriever.from_vg_graph(vg_graph_file_name)
        except OSError:
            retriever = SequenceRetriever.from_vg_json_graph(vg_graph_file_name)

        caller.save_max_path_sequences_to_fasta_file("sequences.fasta", retriever)
    else:
        logging.info("Not saving max path sequences, since a vg graph/sequence retriever was not sent in")


def run_callpeaks_interface(args):
    logging.info("Read offset based graph")

    ob_graph = obg.GraphWithReversals.from_numpy_file(
        args.graph_file_name)

    run_callpeaks(
        ob_graph,
        args.sample_reads_file_name,
        args.control_reads_file_name,
        args.vg_graph_file_name,
        args.out_base_name,
        has_control=args.with_control == "True",
        fragment_length=int(args.fragment_length),
        read_length=int(args.read_length),
        linear_map_file_name=args.linear_map_base_name
    )


def run_callpeaks_whole_genome(args):
    logging.info("Running whole genome.")
    chromosomes = args.chromosomes.split(",")
    graph_file_names = [args.graphs_location + chrom for chrom in chromosomes]
    linear_map_file_names = [args.linear_maps_location + chrom
                             for chrom in chromosomes]
    vg_graphs = [args.vg_graphs_location + chrom + ".vg"
                 for chrom in chromosomes]
    sequence_retrievers = \
        (SequenceRetriever.from_vg_graph(fn) for fn in vg_graphs)

    if args.sample_reads_base_name.endswith(".intervalcollection"):
        sample_file_names = [
            args.sample_reads_base_name.replace("chrom", chrom)
            for chrom in chromosomes]
        control_file_names = [
            args.control_reads_base_name.replace("chrom", chrom)
            for chrom in chromosomes]
    else:
        sample_base_name = args.sample_reads_base_name.replace(".json", "_")
        control_base_name = args.control_reads_base_name.replace(".json", "_")
        sample_file_names = [sample_base_name + chrom + ".json"
                             for chrom in chromosomes]
        control_file_names = [control_base_name + chrom + ".json"
                              for chrom in chromosomes]

    fragment_length = int(args.fragment_length)
    genome_size = int(args.genome_size)
    min_background = int(args.unique_reads) * fragment_length / genome_size
    logging.info(
        "Computed min background signal to be %.3f using fragment length %f, "
        " %d unique reads, and genome size %d" % (int(min_background),
                                                  fragment_length,
                                                  int(args.unique_reads),
                                                  int(min_background)))

    caller = MultipleGraphsCallpeaks(
        chromosomes,
        graph_file_names,
        sample_file_names,
        control_file_names,
        linear_map_file_names,
        int(args.fragment_length),
        int(args.read_length),
        has_control=args.with_control == "True",
        sequence_retrievers=sequence_retrievers,
        out_base_name=args.out_base_name,
        stop_after_p_values=args.stop_after_p_values == "True",
        min_background=min_background
    )
    caller.run()


def run_callpeaks_whole_genome_from_p_values(args):
    logging.info("Running whole genome from p-values.")
    chromosome = args.chromosome
    chromosomes = [chromosome]
    graph_file_names = [args.graphs_location + chrom for chrom in chromosomes]
    vg_graphs = [args.graphs_location + chrom + ".vg" for chrom in chromosomes]
    sequence_retrievers = \
        (SequenceRetriever.from_vg_graph(fn) for fn in vg_graphs)

    caller = MultipleGraphsCallpeaks(
        chromosomes,
        graph_file_names,
        None,
        None,
        None,
        int(args.fragment_length),
        int(args.read_length),
        has_control=args.with_control == "True",
        sequence_retrievers=sequence_retrievers,
        out_base_name=args.out_base_name
    )
    caller.create_joined_q_value_mapping()
    caller.run_from_p_values(only_chromosome=chromosome)



import logging
import numpy as np

from .peakcollection import PeakCollection


class Reporter:
    def __init__(self, base_name):
        self._base_name = base_name

    def sub_graphs(self, data):
        np.savez(self._base_name + "sub_graphs.graphs",
                 **{"peak%s" % i: p._graph for i, p in enumerate(data)})
        np.savez(self._base_name + "sub_graphs.nodeids",
                 **{"peak%s" % i: p._node_ids for i, p in enumerate(data)})

    def qvalues(self, data):
        data.to_sparse_files(
            self._base_name + "qvalues")

    def all_max_paths(self, data):
        PeakCollection(data).to_file(
            self._base_name+"all_max_paths.intervalcollection",
            text_file=True)

    def max_paths(self, data):
        PeakCollection(data).to_file(
            self._base_name+"max_paths.intervalcollection",
            text_file=True)

    def hole_cleaned(self, data):
        data.to_sparse_files(self._base_name+"hole_cleaned")

    def thersholded(self, data):
        data.to_sparse_files(self._base_name+"thresholded")

    def add(self, name, data):
        if hasattr(self, name):
            getattr(self, name)(data)
            logging.info("Wrote %s to file", name)
        else:
            logging.info("Skipping reporting of %s", name)



from graph_peak_caller.peakcollection import PeakCollection
from graph_peak_caller.nongraphpeaks import NonGraphPeakCollection
from offsetbasedgraph import IntervalCollection
from offsetbasedgraph import IndexedInterval, NumpyIndexedInterval
from graph_peak_caller.util import create_linear_path
import pickle
import pyvg
import logging


class AnalysisResults:
    def __init__(self):
        self.tot_peaks1 = 0
        self.tot_peaks2 = 0
        self.peaks1_in_peaks2 = 0
        self.peaks2_in_peaks1 = 0
        self.peaks2_not_in_peaks1 = 0
        self.peaks1_not_in_peaks2 = 0
        self.peaks2_not_in_peaks1_matching_motif = 0
        self.peaks1_not_in_peaks2_matching_motif = 0
        self.peaks1_in_peaks2_matching_motif = 0
        self.peaks2_in_peaks1_matching_motif = 0
        self.motif_ambiguous = 0
        self.motif_not_ambiguous = 0
        self.not_motif_ambiguous = 0
        self.not_motif_not_ambiguous = 0

    def __repr__(self):
        out = ""
        out += "Number of peaks1: %d \n" % self.tot_peaks1
        out += "Number of peaks2: %d \n" % self.tot_peaks2
        out += "Number of peaks1 also found in peaks2: %d (%d with motif hit) \n" % \
               (self.peaks1_in_peaks2, self.peaks1_in_peaks2_matching_motif)
        out += "Number of peaks2 also found in peaks1: %d (%d with motif hit) \n" % \
               (self.peaks2_in_peaks1, self.peaks2_in_peaks1_matching_motif)

        out += "\n"
        out += "Number of peaks1 NOT found in peaks 2: %d (%d with motif hit) \n" % \
               (self.peaks1_not_in_peaks2, self.peaks1_not_in_peaks2_matching_motif)
        out += "Number of peaks2 NOT found in peaks 1: %d (%d with motif hit) \n" % \
               (self.peaks2_not_in_peaks1, self.peaks2_not_in_peaks1_matching_motif)
        out += "\n \n"
        return out

    def __str__(self):
        return self.__repr__()

    def __add__(self, other):
        self.tot_peaks1 += other.tot_peaks1
        self.tot_peaks2 += other.tot_peaks2
        self.peaks1_in_peaks2 += other.peaks1_in_peaks2
        self.peaks2_in_peaks1 += other.peaks2_in_peaks1
        self.peaks2_not_in_peaks1 += other.peaks2_not_in_peaks1
        self.peaks1_not_in_peaks2 += other.peaks1_not_in_peaks2
        self.peaks2_not_in_peaks1_matching_motif += other.peaks2_not_in_peaks1_matching_motif
        self.peaks1_not_in_peaks2_matching_motif += other.peaks1_not_in_peaks2_matching_motif
        self.peaks2_in_peaks1_matching_motif += other.peaks2_in_peaks1_matching_motif
        self.peaks1_in_peaks2_matching_motif += other.peaks1_in_peaks2_matching_motif
        self.motif_ambiguous += other.motif_ambiguous
        self.motif_not_ambiguous += other.motif_not_ambiguous
        self.not_motif_ambiguous += other.not_motif_ambiguous
        self.not_motif_not_ambiguous += other.not_motif_not_ambiguous
        return self

    def to_file(self, file_name):
        with open(file_name, "wb") as f:
            pickle.dump(self, f)

    @classmethod
    def from_file(cls, file_name):
        f = open(file_name, "rb")
        results = pickle.load(f)
        f.close()
        return results

# $ graph_peak_caller analyse_peaks graph.nobg haplo1kg50-mhc.json  macs_sequences_mhc.fasta test_sequences.fasta  fimo_macs_sequences/fimo.txt fimo_test_sequences/fimo.txt chr6 28510119 33480577
# $ graph_peak_caller analyse_peaks ~/dev/graph_peak_caller/tests/whole_genome/22.nobg ~/dev/graph_peak_caller/tests/whole_genome/22.json macs_sequences_chr22.fasta 22_sequences.fasta fimo_macs_chr22/fimo.txt fimo_graph_chr22/fimo.txt 22 22 0 0


class PeaksComparerV2(object):
    def __init__(self, graph,
                 linear_peaks_fasta_file_name,
                 graph_peaks_fasta_file_name,
                 linear_peaks_fimo_results_file,
                 graph_peaks_fimo_results_file,
                 linear_path,
                 region=None):

        assert isinstance(linear_path, NumpyIndexedInterval), \
            "Linear path should be numpy indexed interval for fast comparison"

        print(linear_path)

        self.graph = graph
        self.graph_peaks_fasta_file_name = graph_peaks_fasta_file_name
        self.linear_peaks_fasta_file_name = linear_peaks_fasta_file_name

        self.linear_matching_motif = self._get_peaks_matching_motif(
                linear_peaks_fimo_results_file)
        self.graph_matching_motif = self._get_peaks_matching_motif(
            graph_peaks_fimo_results_file)

        self.linear_path = linear_path
        self.peaks1 = PeakCollection.from_fasta_file(self.graph_peaks_fasta_file_name,
                                                     graph)

        self.peaks1.create_node_index()

        nongraphpeaks = NonGraphPeakCollection.from_fasta(self.linear_peaks_fasta_file_name)

        if region is not None:
            nongraphpeaks.filter_peaks_outside_region(
                region.chromosome, region.start, region.end)
        else:
            logging.info(
                "Graph region is None. Whole graph will be used and no filtering on peaks will be done.")

        self.peaks2 = PeakCollection.create_from_nongraph_peak_collection(
            graph,
            nongraphpeaks,
            self.linear_path,
            graph_region=region)

        print("%d linear peaks converted to graph" % len(self.peaks2.intervals))

        self.peaks2.create_node_index()

        self.results = AnalysisResults()

        self.peaks1_in_peaks2 = []
        self.peaks2_in_peaks1 = []
        self.peaks1_not_in_peaks2 = []
        self.peaks2_not_in_peaks1 = []
        self.run_all_analysis()

    def run_all_analysis(self):
        self.check_similarity()
        self.check_non_matching_for_motif_hits()
        self.check_matching_for_motif_hits()

        print(self.results)

    def check_matching_for_motif_hits(self):
        print("\n--- Checking peaks matching for motif hits --- ")
        self._get_pair_ambig_status()
        i = 1
        for matching in [self.peaks1_in_peaks2, self.peaks2_in_peaks1]:
            if i == 1:
                print("Checking graph peaks, in total %d in other set" % len(matching))
                matching_motif = self.graph_matching_motif
            else:
                print("Checking linear peaks, in total %d in other set" % len(matching))
                matching_motif = self.linear_matching_motif

            a, b, c, d = self.count_matching_motif(matching_motif, matching, with_info=True)
            n = a + b
            print("Found n peaks that matches motif: %d" % n)

            if i == 1:
                self.results.motif_ambiguous = a
                self.results.motif_not_ambiguous = b
                self.results.not_motif_ambiguous = c
                self.results.not_motif_not_ambiguous = d
                self.results.peaks1_in_peaks2_matching_motif = n
            else:
                print("MA: ", a, "MNA:", b, "NMA:", c, "NMNA:", d)
                self.results.peaks2_in_peaks1_matching_motif = n
            i += 1

    def check_non_matching_for_motif_hits(self):
        print("\n--- Checking peaks that are not in other set for motif hits --- ")
        i = 1
        for not_matching in [self.peaks1_not_in_peaks2, self.peaks2_not_in_peaks1]:
            if i == 1:
                print("Graph peaks")
                matching_motif = self.graph_matching_motif
            else:
                print("Linear peaks")
                matching_motif = self.linear_matching_motif

            n = self.count_matching_motif(matching_motif, not_matching)
            print("N matching motif: %d" % n)

            if i == 1:
                self.results.peaks1_not_in_peaks2_matching_motif = n
            else:
                self.results.peaks2_not_in_peaks1_matching_motif = n

            i += 1

    def _get_pair_ambig_status(self):
        peaks = self.peaks1_in_peaks2
        matches = [self.peaks2.which_approx_contains_part_of_interval(peak)
                   for peak in peaks]
        my_motif_matches = [peak.unique_id in self.graph_matching_motif
                            for peak in peaks]
        other_motif_matches = [peak.unique_id in self.linear_matching_motif
                               for peak in peaks]
        is_ambiguous = [peak.info[0] for peak in peaks]
        n = 0
        found = 0
        for my_match, other_match, ambig in zip(my_motif_matches, other_motif_matches, is_ambiguous):
            if other_match and not my_match:
                n += 1
                if ambig:
                    found += 1
        print("Ambigous of non-found: ", found, n)

    def count_matching_motif(self, matching_motif_ids, peak_list, with_info=False):
        a, b, c, d = (0, 0, 0, 0)
        for peak in peak_list:
            assert peak.unique_id is not None
            ambig = peak.info[0]
            motif = peak.unique_id in matching_motif_ids
            if motif and ambig:
                a += 1
            elif motif and not ambig:
                b += 1
            elif (not motif) and ambig:
                c += 1
            else:
                d += 1
        if with_info:
            return a, b, c, d
        return a+b

    def _get_peaks_matching_motif(self, fimo_file_name):
        matches = set()
        f = open(fimo_file_name)
        for line in f:
            l = line.split()
            sequence_id = l[2]
            matches.add(sequence_id)
        f.close()
        return matches

    def _read_peaks(self):
        self.peaks1 = PeakCollection.from_fasta_file(self.graph_peaks_fasta_file_name)

    def plot_peak_lengths(self):
        import matplotlib.pyplot as plt
        i = 1
        for peak_set in (self.peaks1, self.peaks2):
            plt.hist([peak.length() for peak in peak_set],
                     bins=1000, label="Peak set %d" % i)
            i += 1

        plt.legend()
        plt.show()

    def get_peaks_at_same_position(self, allowed_mismatches=0):
        same_pos = []
        for peak in self.peaks1:
            identical = self.peaks2.get_similar_intervals(
                peak, allowed_mismatches)

            assert len(identical) <= 1
            if len(identical) > 0:
                same_pos.append((peak, identical[0]))

        return same_pos

    @classmethod
    def create_from_graph_peaks_and_linear_peaks(
            cls,
            linear_peaks_bed_file_name,
            graph_peaks_file_name,
            ob_graph,
            linear_path,
            graph_region=None):
        linear_path = linear_path.to_indexed_interval()
        linear_peaks = PeakCollection.create_from_linear_intervals_in_bed_file(
            ob_graph, linear_path, linear_peaks_bed_file_name,
            graph_region=graph_region)

        linear_peaks.to_file("macs_peaks.intervals", text_file=True)
        sequence_retriever = None
        comparer = PeaksComparer(ob_graph, sequence_retriever,
                                 None,
                                 "macs_peaks.intervals",
                                 graph_peaks_file_name)
        return comparer

    def check_similarity(self, analyse_first_n_peaks=10000000):
        i = 1
        for peak_datasets in [(self.peaks1, self.peaks2),
                              (self.peaks2, self.peaks1)]:
            n_identical = 0
            tot_n_similar = 0
            n_similar = 0
            n_tot = 0
            print("\n-- Comparing set %d against set %d ---" % (i, i % 2 + 1))
            peaks1, peaks2 = peak_datasets
            print("Number of peaks in main set: %d" % len(peaks1.intervals))
            if i == 1:
                self.results.tot_peaks1 = len(peaks1.intervals)
            else:
                self.results.tot_peaks2 = len(peaks1.intervals)

            not_matching = []
            matching = []
            counter = 0
            for peak in sorted(peaks1, key=lambda x: x.score, reverse=True)[0:analyse_first_n_peaks]:
                assert peak.unique_id is not None
                counter += 1
                #if peaks2.contains_interval(peak):
                #    n_identical += 1

                if counter % 500 == 0:
                    logging.info("Checked %d peaks" % counter)


                #similar_intervals = peaks2.get_overlapping_intervals(peak, 50)
                #print("Peak %s is overlapping with %d other peaks" % (peak.unique_id, len(similar_intervals)))
                #if len(similar_intervals) > 0:
                if peaks2.approx_contains_part_of_interval(peak):
                    n_similar += 1
                    if i == 1:
                        self.peaks1_in_peaks2.append(peak)
                    else:
                        self.peaks2_in_peaks1.append(peak)

                    matching.append(peak)
                else:
                    not_matching.append(peak)
                    if i == 1:
                        self.peaks1_not_in_peaks2.append(peak)
                    else:
                        self.peaks2_not_in_peaks1.append(peak)

                n_tot += 1
            self.results.peaks1_in_peaks2 = len(self.peaks1_in_peaks2)
            self.results.peaks2_in_peaks1 = len(self.peaks2_in_peaks1)

            self.results.peaks1_not_in_peaks2 = len(self.peaks1_not_in_peaks2)
            self.results.peaks2_not_in_peaks1 = len(self.peaks2_not_in_peaks1)

            not_matching = IntervalCollection(not_matching)
            not_matching.to_file("not_matching_set%d.intervals" % i, text_file=True)
            logging.info("Wrote peaks not matching to file not_matching_set%d.intervals" % i)
            matching = IntervalCollection(matching)
            matching.to_file("matching_set%d.intervals" % i, text_file=True)

            logging.info("Total peaks in main set: %d" % n_tot)
            logging.info("N similar to peak in other set: %d " % n_similar)
            logging.info("N not matching other set: %d " % len(not_matching.intervals))

            i += 1

    def check_overlap_with_linear_path(self):
        i = 0
        for peaks in [self.peaks1, self.peaks2]:
            print("\n -- Checking peak dataset % against linear path -- " % (i))
            i += 1
            n_inside = 0
            n_inside_correct_order = 0
            for peak in peaks:
                if self.linear_path.contains(peak):
                    n_inside += 1
                if self.linear_path.contains_in_order_any_direction(peak):
                    n_inside_correct_order += 1

            print("n inside: %d / %d" % (n_inside, len(peaks.intervals)))
            print("n inside correct order: %d / %d" % (
                n_inside_correct_order, len(peaks.intervals)))

    def get_peaks_on_linear_path(self, peaks):
        out = []
        for peak in peaks:
            if self.linear_path.contains_in_correct_order(peak):
                out.append(peak)
        return out

    def get_peaks_not_on_linear_path(self):
        out = []
        for peak in self.peaks2:
            if not self.linear_path.contains_in_correct_order(peak):
                out.append(peak)
        return out

    def peaks_to_fasta(self, peaks, file_name="peaks.fasta"):
        f = open(file_name, "w")
        i = 1
        for p in peaks:
            sequence = self.sequence_retriever.get_interval_sequence(p)
            f.writelines([">peak%d\n%s\n" % (i, sequence)])
            i += 1
        f.close()

    def graph_peaks_on_main_path_not_in_linear(self):
        on_linear = self.get_peaks_on_linear_path(self.peaks2)
        print("On linear path: %d" % len(on_linear))
        on_linear = PeakCollection(on_linear)
        return on_linear.get_peaks_not_in_other_collection(self.peaks1, 20)


class PeaksComparer(object):

    def __init__(self, graph, sequence_retriever, linear_path_file_name,
                 peaks1_file_name, peaks2_file_name):
        self.graph = graph
        self.sequence_retriever = sequence_retriever
        self.peaks1 = PeakCollection.create_list_from_file(
            peaks1_file_name, graph=graph)
        self.peaks2 = PeakCollection.create_list_from_file(
            peaks2_file_name, graph=graph)
        print("Number of intervals in set 1/2: %d / %d" % (
            len(self.peaks1.intervals), len(self.peaks2.intervals)))
        if linear_path_file_name is not None:
            self.linear_path = IntervalCollection.create_list_from_file(
                linear_path_file_name, self.graph).intervals[0]

    def plot_peak_lengths(self):
        import matplotlib.pyplot as plt
        i = 1
        for peak_set in (self.peaks1, self.peaks2):
            plt.hist([peak.length() for peak in peak_set],
                     bins=1000, label="Peak set %d" % i)
            i += 1

        plt.legend()
        plt.show()

    def get_peaks_at_same_position(self, allowed_mismatches=0):
        same_pos = []
        for peak in self.peaks1:
            identical = self.peaks2.get_similar_intervals(
                peak, allowed_mismatches)

            assert len(identical) <= 1
            if len(identical) > 0:
                same_pos.append((peak, identical[0]))

        return same_pos

    def compare_q_values_for_similar_peaks(self):

        for peak in self.peaks1:
            similar = self.peaks2.get_similar_intervals(
                peak, allowed_mismatches=10)
            if len(similar) > 0:
                print("Found match(es) for %s" % peak)
                for matched_peak in similar:
                    print("   Match agsinst %s with scores %.3f, %.3f" %
                          (matched_peak, peak.score, matched_peak.score))
            else:
                print("No match for peak %s" % peak)

    @classmethod
    def create_from_graph_peaks_and_linear_peaks(
            cls,
            linear_peaks_bed_file_name,
            graph_peaks_file_name,
            ob_graph,
            linear_path,
            graph_region=None):
        linear_path = linear_path.to_indexed_interval()
        linear_peaks = PeakCollection.create_from_linear_intervals_in_bed_file(
            ob_graph, linear_path, linear_peaks_bed_file_name,
            graph_region=graph_region)

        linear_peaks.to_file("macs_peaks.intervals", text_file=True)
        sequence_retriever = None
        comparer = PeaksComparer(ob_graph, sequence_retriever,
                                 None,
                                 "macs_peaks.intervals",
                                 graph_peaks_file_name)
        return comparer

    def check_similarity(self, analyse_first_n_peaks=250):
        i = 1
        for peak_datasets in [(self.peaks1, self.peaks2),
                              (self.peaks2, self.peaks1)]:
            n_identical = 0
            tot_n_similar = 0
            n_similar = 0
            n_tot = 0
            print("\n-- Comparing set %d against set %d ---" % (i, i % 2 + 1))
            peaks1, peaks2 = peak_datasets
            print("Number of peaks in main set: %d" % len(peaks1.intervals))
            not_matching = []
            matching = []
            counter = 0
            for peak in sorted(peaks1, key=lambda x: x.score, reverse=True)[0:analyse_first_n_peaks]:
                counter += 1
                if peaks2.contains_interval(peak):
                    n_identical += 1

                similar_intervals = peaks2.get_overlapping_intervals(peak, 50)
                if len(similar_intervals) > 0:
                    n_similar += 1
                    tot_n_similar += len(similar_intervals)
                    print(counter, peak.score, peak.start_position)
                    matching.append(peak)
                else:
                    not_matching.append(peak)
                    print(peak, "\t", 0)

                n_tot += 1

            not_matching = PeakCollection(not_matching)
            not_matching.to_file("not_matching_set%d.intervals" % i, text_file=True)
            matching = PeakCollection(matching)
            matching.to_file("matching_set%d.intervals" % i, text_file=True)

            print("Total peaks in main set: %d" % n_tot)
            print("N identical to peak in other set: %d " % n_identical)
            print("N similar to peak in other set: %d " % n_similar)
            print("Total number of similar hits (counting all hits for each peaks): %d " % tot_n_similar)

            i += 1

    def check_overlap_with_linear_path(self):
        i = 0
        for peaks in [self.peaks1, self.peaks2]:
            print("\n -- Checking peak dataset % against linear path -- " % (i))
            i += 1
            n_inside = 0
            n_inside_correct_order = 0
            for peak in peaks:
                if self.linear_path.contains(peak):
                    n_inside += 1
                if self.linear_path.contains_in_order_any_direction(peak):
                    n_inside_correct_order += 1

            print("n inside: %d / %d" % (n_inside, len(peaks.intervals)))
            print("n inside correct order: %d / %d" % (
                n_inside_correct_order, len(peaks.intervals)))

    def get_peaks_on_linear_path(self, peaks):
        out = []
        for peak in peaks:
            if self.linear_path.contains_in_correct_order(peak):
                out.append(peak)
        return out

    def get_peaks_not_on_linear_path(self):
        out = []
        for peak in self.peaks2:
            if not self.linear_path.contains_in_correct_order(peak):
                out.append(peak)
        return out

    def peaks_to_fasta(self, peaks, file_name="peaks.fasta"):
        f = open(file_name, "w")
        i = 1
        for p in peaks:
            sequence = self.sequence_retriever.get_interval_sequence(p)
            f.writelines([">peak%d\n%s\n" % (i, sequence)])
            i += 1
        f.close()

    def graph_peaks_on_main_path_not_in_linear(self):
        on_linear = self.get_peaks_on_linear_path(self.peaks2)
        print("On linear path: %d" % len(on_linear))
        on_linear = PeakCollection(on_linear)
        return on_linear.get_peaks_not_in_other_collection(self.peaks1, 20)



import numpy as np
import logging
from itertools import chain
from collections import defaultdict

import offsetbasedgraph as obg

from ..sparsediffs import SparseDiffs
from ..densepileup import DensePileup


class NodeInfo:
    def __init__(self, dist_dict=None):
        self._dist_dict = dist_dict if dist_dict\
                          is not None else defaultdict(int)

    def update(self, starts):
        for read_id, val in starts.items():
            self._dist_dict[read_id] = max(self._dist_dict[read_id], val)

    def __str__(self):
        return str(self._dist_dict)

    def __repr__(self):
        return str(self._dist_dict)


class SparseGraphPileup:
    def __init__(self, graph):
        self.starts = []
        self.ends = []
        self.node_starts = np.zeros(len(graph.blocks)+2)
        self.touched_nodes = np.zeros(self.node_starts.size, dtype="bool")

    def __str__(self):
        return("\n".join([str(self.starts), str(self.ends), str(self.node_starts)]))


class ReadsAdder:
    def __init__(self, graph, pileup, keep_ends=False):
        self.min_id = graph.min_node
        self._node_indexes = graph.node_indexes
        self._pileup = pileup
        self._pos_ends = {node_id: [] for node_id in graph.blocks.keys()}
        self._neg_ends = {-node_id: [] for node_id
                          in graph.blocks.keys()}

    def _handle_interval(self, interval):
        # self._pileup.add_interval(interval)
        end_pos = interval.end_position
        rp = end_pos.region_path_id
        if rp < 0:
            self.add_open_neg_interval(interval)
            self._neg_ends[rp].append(end_pos.offset)
        else:
            self.add_open_pos_interval(interval)
            self._pos_ends[rp].append(end_pos.offset)

    def get_pos_ends(self):
        return self._pos_ends

    def get_neg_ends(self):
        return self._neg_ends

    def _add_start_pos(self, pos):
        rp = (abs(pos.region_path_id)-self.min_id)
        if pos.region_path_id > 0:
            self._pileup.starts.append(self._node_indexes[rp]+pos.offset)
        else:
            self._pileup.ends.append(self._node_indexes[rp+1]-pos.offset)

    def add_open_pos_interval(self, interval):
        rps = [rp-self.min_id for rp in interval.region_paths]
        self._pileup.touched_nodes[rps] = True
        self._pileup.node_starts[rps[1:]] += 1
        self._pileup.node_starts[1:][rps[:-1]] -= 1
        self._add_start_pos(interval.start_position)

    def add_open_neg_interval(self, interval):
        rps = [abs(rp)-self.min_id for rp in interval.region_paths]
        self._pileup.touched_nodes[rps] = True
        self._pileup.node_starts[1:][rps[1:]] -= 1
        self._pileup.node_starts[rps[:-1]] += 1
        self._add_start_pos(interval.start_position)

    def add_reads(self, intervals):
        i = 0
        for interval in intervals:
            if i % 5000 == 0:
                logging.info("%d reads processed" % i)
            self._handle_interval(interval)
            i += 1


class ReadsAdderWDirect(ReadsAdder):
    def __init__(self, graph, pileup):
        self.pos_read_ends = []
        self.neg_read_ends = []
        super().__init__(graph, pileup)

    def _handle_interval(self, interval):
        # self._pileup.add_interval(interval)
        super()._handle_interval(interval)
        end_pos = interval.end_position
        rp = end_pos.region_path_id
        if rp < 0:
            self.neg_read_ends.append(
                self._node_indexes[abs(rp)+1-self.min_id]-end_pos.offset)
        else:
            self.pos_read_ends.append(
                self._node_indexes[abs(rp)-self.min_id]+end_pos.offset)


class SparseExtender:
    def __init__(self, graph, pileup, fragment_length):
        self._graph = graph
        self._pileup = pileup
        self._fragment_length = fragment_length
        self._set_adj_list()
        self._graph_size = graph.node_indexes[-1]

    def _set_adj_list(self):
        self._adj_list = self._graph.adj_list

    def get_node_ids(self):
        return range(self._graph.min_node, self._graph._id+1)
    # return self._graph.get_sorted_node_ids()

    def _add_node_end(self, node_id, value):
        self._pileup.node_starts[node_id+1-self._graph.min_node] -= value

    def _add_node_start(self, node_id, value):
        self._pileup.node_starts[node_id-self._graph.min_node] += value

    def _add_end(self, index):
        self._pileup.ends.append(index)

    def run_linear(self, starts_dict):
        node_ids = self.get_node_ids()
        node_infos = defaultdict(NodeInfo)
        cur_id = 0
        empty = NodeInfo()
        cur_array_idx = 0
        n_nodes = len(node_ids)
        for i, node_id in enumerate(node_ids):
            if i % 1000000 == 0:
                logging.info("Handling node %s of %s", i, n_nodes)
            info = node_infos.pop(node_id, empty)
            if info._dist_dict:
                self._pileup.touched_nodes[
                    abs(node_id)-self._graph.min_node] = True
            node_size = self._graph.node_size(node_id)
            starts = starts_dict[node_id]
            n_starts = len(starts)
            # n_ends = n_starts + len(info._dist_dict)
            endsidxs = chain(enumerate(
                (start+self._fragment_length for start in starts),
                start=cur_id),
                             info._dist_dict.items())
            self._add_node_start(node_id, len(info._dist_dict))
            d = {}
            for idx, end in endsidxs:
                if end <= node_size:
                    self._add_end(cur_array_idx+end)
                else:
                    d[idx] = end-node_size
            cur_array_idx += node_size
            cur_id = cur_id + n_starts
            self._add_node_end(node_id, len(d))
            for next_node in self._adj_list[node_id]:
                node_infos[next_node].update(d)

    def get_pileup(self):
        return np.cumsum(self._pileup)


class ReverseSparseExtender(SparseExtender):
    def get_pileup(self):
        return np.cumsum(self._pileup[:-1])[::-1]

    def _set_adj_list(self):
        self._adj_list = self._graph.reverse_adj_list

    def get_node_ids(self):
        return range(-self._graph._id, -self._graph.min_node+1)
    # return (-node_id for node_id in
    #             self._graph.get_sorted_node_ids(reverse=True))

    def _add_node_end(self, node_id, value):
        self._pileup.node_starts[-node_id-self._graph.min_node] += value

    def _add_node_start(self, node_id, value):
        self._pileup.node_starts[-node_id+1-self._graph.min_node] -= value

    def _add_end(self, index):
        self._pileup.starts.append(self._graph_size-index)


def _get_node_indexes(graph):
    if isinstance(graph.blocks, obg.BlockArray):
        # Quicker way to make node_indexes array
        logging.info("(using cumsum on np block array)")
        node_indexes = np.cumsum(graph.blocks._array, dtype=np.uint32)
        logging.info("Node indexes created...")
        graph.min_node = (graph.blocks.node_id_offset+1)
        return node_indexes
    sorted_nodes = sorted(graph.blocks.keys())
    min_node = sorted_nodes[0]
    graph.min_node = min_node
    max_node = sorted_nodes[-1]
    span = max_node-min_node+1
    node_indexes = np.zeros(span+1, dtype=np.uint32)
    offset = 0
    for i, node in enumerate(sorted_nodes):
        index = node - min_node
        node_indexes[index] = offset
        offset += graph.node_size(node)
        node_indexes[-1] = offset
    return node_indexes


class SamplePileupGenerator:
    def __init__(self, graph, extension):
        graph.node_indexes = _get_node_indexes(graph)
        self._pileup = SparseGraphPileup(graph)
        self._graph = graph
        self._reads_adder = ReadsAdderWDirect(graph, self._pileup)
        self._pos_extender = SparseExtender(graph, self._pileup, extension)
        self._neg_extender = ReverseSparseExtender(
            graph, self._pileup, extension)

    def __to_dense(self):
        diffs = SparseDiffs.from_pileup(self._pileup,
                                        self._graph.node_indexes)
        sparse_values = diffs.get_sparse_values()
        densepileup = DensePileup(self._graph)
        for start, end, value in zip(
                sparse_values.indices,
                chain(sparse_values.indices[1:], [densepileup.data._values.size]),
                sparse_values.values):
            densepileup.data._values[start:end] = value
        densepileup.data._touched_nodes = {
            node_id for node_id in self._graph.blocks if
            np.count_nonzero(densepileup.data.values(node_id))}
        return densepileup

    def save_direct(self, base_name):
        new_pileup = SparseGraphPileup(self._graph)
        new_pileup.ends = self._pileup.ends + self._reads_adder.pos_read_ends
        new_pileup.starts = self._pileup.starts + self._reads_adder.neg_read_ends
        new_pileup.node_starts = self._pileup.node_starts
        sparsediff = SparseDiffs.from_pileup(
            new_pileup, self._graph.node_indexes)
        sparsediff.clean()
        sparse_values = sparsediff.get_sparse_values()
        sparse_values.track_size = self._graph.node_indexes[-1]
        sparse_values.to_sparse_files(base_name)
        # np.save(base_name + "_diffindices.npy", sparsediff._indices)
        # np.save(base_name + "_diffvalues.npy", sparsediff._diffs)

    def run(self, reads, save_name=None):
        self._reads_adder.add_reads(reads)
        if save_name is not None:
            self.save_direct(save_name)
        self._pos_extender.run_linear(self._reads_adder.get_pos_ends())
        self._neg_extender.run_linear(self._reads_adder.get_neg_ends())
        sdiffs = SparseDiffs.from_pileup(self._pileup,
                                         self._graph.node_indexes)
        sdiffs.touched_nodes = set(
            np.flatnonzero(
                self._pileup.touched_nodes[:-2]) + self._graph.min_node)
        return sdiffs
    # return self.__to_dense()



import logging
import numpy as np
import scipy.sparse.csgraph as csgraph
from scipy.sparse import csr_matrix
from itertools import chain

from .segmentanalyzer import SegmentAnalyzer
from ..sparsediffs import SparseValues
from .subgraphanalyzer import SubGraphAnalyzer


class DummyTouched:
    def __contains__(self, item):
        return True


class SubGraph:
    def __init__(self, node_ids, graph):
        self._node_ids = node_ids
        self._graph = graph


class StubsFilter:
    def __init__(self, starts, fulls, ends, graph,
                 last_node=None, touched_nodes=None):
        self._last_node = last_node
        self._graph = graph
        self._starts = starts
        self._fulls = fulls
        self._ends = ends

        self._touched_nodes = touched_nodes
        if touched_nodes is None:
            self._touched_nodes = DummyTouched()

        self._starts_mask = np.ones_like(starts, dtype="bool")
        self._fulls_mask = np.ones_like(fulls, dtype="bool")
        self._ends_mask = np.ones_like(ends, dtype="bool")

        self.filter_start_stubs()
        self.filter_end_stubs()
        self.filtered_ends = self._ends[self._ends_mask]
        self.filtered_starts = self._starts[self._starts_mask]
        self.filtered_fulls = self._fulls[self._fulls_mask]
        self._set_pos_nodes()

        self._full_starts = np.flatnonzero(
            self.find_sub_starts(self.filtered_fulls))
        self._end_starts = np.flatnonzero(
            self.find_sub_starts(self.filtered_ends))
        self._full_ends = np.flatnonzero(
            self.find_sub_ends(self.filtered_fulls))
        self._start_ends = np.flatnonzero(
            self.find_sub_ends(self.filtered_starts))

    def find_sub_starts(self, nodes):
        r = np.array([not all(-adj in self._pos_from_nodes or adj
                              not in self._touched_nodes
                              for adj in self._graph.reverse_adj_list[-node])
                      for node in nodes], dtype="bool")
        return r

    def find_sub_ends(self, nodes):
        a = np.array([not all(adj in self._pos_to_nodes or adj > self._last_node or adj not in self._touched_nodes
                              for adj in self._graph.adj_list[node])
                      for node in nodes], dtype="bool")
        return a

    def _get_start_filter(self, nodes):
        return np.array([len(self._graph.reverse_adj_list[-node_id])
                         for node_id in nodes], dtype="bool")

    def _get_ends_filter(self, nodes):
        return np.array([len(self._graph.adj_list[node_id])
                         for node_id in nodes], dtype="bool")

    def filter_start_stubs(self):
        """ Locate nodes that are start_nodes of graph"""
        self._fulls_mask &= self._get_start_filter(self._fulls)
        self._ends_mask &= self._get_start_filter(self._ends)

    def filter_end_stubs(self):
        self._starts_mask &= self._get_ends_filter(self._starts)
        self._fulls_mask &= self._get_ends_filter(self._fulls)

    def _set_pos_nodes(self):
        self._pos_to_nodes = set(chain(self._fulls, self._ends))
        self._pos_from_nodes = set(chain(self._starts, self._fulls))


class PosStubFilter(StubsFilter):

    def _set_pos_nodes(self):
        self._pos_to_nodes = set(chain(self._fulls, self._ends))
        self._pos_from_nodes = set(chain(self._starts, self._fulls))

    def find_sub_starts(self, nodes):
        return np.array([not any(-adj in self._pos_from_nodes
                                 for adj in self._graph.reverse_adj_list[-int(node)])
                         for node in nodes], dtype="bool")

    def find_sub_ends(self, nodes):
        return np.array([not any(adj in self._pos_to_nodes
                                 for adj in self._graph.adj_list[int(node)])
                         for node in nodes], dtype="bool")

    def filter_start_stubs(self):
        """ Locate nodes that are start_nodes of graph"""
        pass

    def filter_end_stubs(self):
        pass


class LineGraph:
    stub_class = StubsFilter
    def __init__(self, starts, full, ends, ob_graph, last_node=None, touched_nodes=None):
        if last_node is not None:
            last_node += ob_graph.min_node-1
        starts[0] += ob_graph.min_node-1
        full[0] += ob_graph.min_node-1
        ends[0] += ob_graph.min_node-1
        self.filtered = self.stub_class(starts[0], full[0], ends[0],
                                        ob_graph, last_node, touched_nodes)
        self.full_size = full[1][self.filtered._fulls_mask]
        self.start_size = starts[1][self.filtered._starts_mask]
        self.end_size = ends[1][self.filtered._ends_mask]

        self.kept_starts = np.vstack((starts[0][~self.filtered._starts_mask],
                                      starts[1][~self.filtered._starts_mask]))
        self.kept_ends = np.vstack((ends[0][~self.filtered._ends_mask],
                                    ends[1][~self.filtered._ends_mask]))
        self.kept_fulls = full[0][~self.filtered._fulls_mask]
        self.kept_fulls = self.kept_fulls.astype("int")

        self.full_nodes = self.filtered.filtered_fulls
        self.start_nodes = self.filtered.filtered_starts
        self.end_nodes = self.filtered.filtered_ends
        self._all_nodes = np.r_[self.start_nodes,
                                self.full_nodes,
                                self.end_nodes]
        self._all_nodes = self._all_nodes.astype("int")
        self._all_sizes = np.r_[self.start_size,
                                self.full_size,
                                self.end_size]
        self._all_sizes = self._all_sizes.astype("int")
        self.ob_graph = ob_graph

        self.n_starts = self.start_nodes.size
        self.n_ends = self.end_nodes.size
        self.n_nodes = self._all_nodes.size
        self._matrix = self.make_graph()

    def make_graph(self):
        n_starts = self.start_nodes.size
        n_ends = self.end_nodes.size
        to_nodes_dict = {node: n_starts+i for i, node in
                         enumerate(self._all_nodes[n_starts:])}
        self.end_stub = self._all_nodes.size
        from_nodes = []
        to_nodes = []
        sizes = []
        possible_to_nodes = set(list(self._all_nodes[n_starts:]))
        for i, node_id in enumerate(self._all_nodes[:self._all_nodes.size-n_ends]):
            adj_nodes = self.ob_graph.adj_list[node_id]
            next_nodes = [to_nodes_dict[node] for node in
                          adj_nodes if node in possible_to_nodes]
            from_nodes.extend([i]*len(next_nodes))
            to_nodes.extend(next_nodes)
            sizes.extend([self._all_sizes[i]]*len(next_nodes))

        end_nodes = np.r_[self.filtered._start_ends,
                          n_starts + self.filtered._full_ends,
                          np.arange(self.n_nodes-n_ends, self.n_nodes)]
        to_nodes.extend([self.end_stub]*end_nodes.size)
        from_nodes.extend(list(end_nodes))
        sizes.extend(self._all_sizes[end_nodes])
        # self.n_starts = n_starts
        self._graph_node_sizes = sizes
        return csr_matrix((np.array(sizes, dtype="int32"),
                           (np.array(from_nodes, dtype="int32"),
                            np.array(to_nodes, dtype="int32"))),
                          [self.end_stub+1, self.end_stub+1])

    def get_masked(self, mask):
        n_starts, n_ends = self.n_starts, self.n_ends
        self._all_nodes -= (self.ob_graph.min_node-1)
        T = self._all_nodes.size
        starts = np.vstack((self._all_nodes[:n_starts][mask[:n_starts]],
                            self._all_sizes[:n_starts][mask[:n_starts]]))
        fulls = self._all_nodes[n_starts:T-n_ends][mask[n_starts:T-n_ends]]
        ends = np.vstack((self._all_nodes[T-n_ends:][mask[T-n_ends:]],
                          self._all_sizes[T-n_ends:][mask[T-n_ends:]]))
        d = self.ob_graph.min_node-1
        self.kept_starts[0] -= d
        self.kept_ends[0] -= d
        self.kept_fulls -= d
        starts = np.hstack((starts, self.kept_starts))
        ends = np.hstack((ends, self.kept_ends))
        fulls = np.r_[fulls, self.kept_fulls]

        return starts, fulls, ends

    def filter_small(self, max_size):
        start_nodes = np.r_[np.arange(self.n_starts),
                            self.n_starts+self.filtered._full_starts,
                            self.n_nodes-self.n_ends+self.filtered._end_starts]
        if not start_nodes.size:
            return np.array([], dtype="bool")
        connected_components = csgraph.connected_components(self._matrix)
        shortest_paths = csgraph.shortest_path(self._matrix)
        to_dist = np.min(shortest_paths[start_nodes], axis=0)
        from_dist = shortest_paths[:, self.end_stub]
        return (to_dist+from_dist)[:-1] > max_size


class DividedLinegraph(LineGraph):

    def _get_subgraph(self, idxs):
        self._lookup[idxs] = np.arange(idxs.size)
        sub_matrix = self._matrix[idxs]
        new_indices = self._lookup[sub_matrix.indices]
        sub_matrix = csr_matrix((sub_matrix.data,
                                 new_indices,
                                 sub_matrix.indptr),
                                shape=(idxs.size, idxs.size))
        return sub_matrix

    def filter_small(self, max_size):
        start_nodes = np.r_[np.arange(self.n_starts),
                            self.n_starts+self.filtered._full_starts,
                            self.n_nodes-self.n_ends+self.filtered._end_starts]
        if not start_nodes.size:
            return np.ones_like(self._all_nodes, dtype="bool")
        start_nodes_mask = np.zeros(self.end_stub, dtype="bool")
        start_nodes_mask[start_nodes] = True
        if not start_nodes.size:
            return np.array([], dtype="bool")
        n_components, connected_components = csgraph.connected_components(
            self._matrix[:self.end_stub, :self.end_stub])
        logging.info("Found %s components", n_components)
        complete_mask = np.zeros(self.end_stub, dtype="bool")
        self._lookup = np.empty(self.end_stub+1, dtype="int")
        for comp in range(n_components):
            if comp % 100 == 0:
                logging.info("Component %s of %s", comp, n_components)

            idxs = np.r_[np.flatnonzero(
                connected_components == comp), self.end_stub]

            if idxs.size-1 > 36:
                complete_mask[idxs[:-1]] = True
                continue
            my_mask = start_nodes_mask[idxs[:-1]]
            start_nodes = np.flatnonzero(my_mask)
            if not start_nodes.size:
                complete_mask[idxs[:-1]] = True
                continue
            subgraph = self._get_subgraph(idxs)
            shortest_paths = csgraph.shortest_path(subgraph)
            to_dist = np.min(shortest_paths[start_nodes], axis=0)
            from_dist = shortest_paths[:, -1]
            complete_mask[idxs[:-1]] = (to_dist+from_dist)[:-1] > max_size
        return complete_mask


class PosDividedLineGraph(DividedLinegraph):
    stub_class = PosStubFilter

    def _backtrace(self, dists, predecessors, start_nodes):
        start = start_nodes[np.argmin(dists[start_nodes, -1])]
        max_row = predecessors[start]
        cur = max_row[-1]
        path = []
        while cur > 0:
            path.append(cur)
            cur = max_row[cur]
        if (not path) or path[-1] != start:
            path += [start]
        return path

    def max_paths(self):
        start_nodes = np.r_[np.arange(self.n_starts),
                            self.n_starts+self.filtered._full_starts,
                            self.n_nodes-self.n_ends+self.filtered._end_starts]
        if not start_nodes.size:
            return [], [], []
        start_nodes_mask = np.zeros(self.end_stub, dtype="bool")
        start_nodes_mask[start_nodes] = True
        paths = []
        infos = []
        self._matrix.data += 1
        subgraphs = []
        n_components, connected_components = csgraph.connected_components(
            self._matrix[:self.end_stub, :self.end_stub])
        logging.info("Found %s components", n_components)
        # node_id_refs = []
        # subgraphs = []
        for comp in range(n_components):
            if comp % 100 == 0:
                logging.info("Component %s of %s", comp, n_components)
            idxs = np.r_[np.flatnonzero(
                connected_components == comp), self.end_stub]
            subgraph = -self._matrix[idxs][:, idxs]
            subgraph.data += 1
            subgraphs.append(SubGraph(self._all_nodes[idxs[:-1]], subgraph))
            if idxs.size == 2:
                paths.append([idxs[0]])
                infos.append((0, 0))
                continue
            distances, predecessors = csgraph.shortest_path(
                subgraph, return_predecessors=True, method="BF")
            my_mask = start_nodes_mask[idxs[:-1]]
            start_nodes = np.flatnonzero(my_mask)
            local_idxs = self._backtrace(distances, predecessors, start_nodes)
            global_idxs = idxs[local_idxs]
            paths.append(global_idxs[::-1])
            infos.append(
                SubGraphAnalyzer(subgraph, local_idxs[::-1], distances).get_info())
        return paths, infos, subgraphs


class HolesCleaner:
    def __init__(self, graph, sparse_values, max_size, touched_nodes=None):
        sparse_values.indices = sparse_values.indices.astype("int")
        self._graph = graph
        self._node_indexes = graph.node_indexes
        self._sparse_values = sparse_values
        self._holes = self.get_holes()
        self._node_ids = self.get_node_ids()
        self._max_size = max_size
        self._kept = []
        self._touched_nodes = touched_nodes
        if touched_nodes is None:
            self._touched_nodes = DummyTouched()

    def get_holes(self):
        start_idx = 0
        if self._sparse_values.values[0] != 0:
            start_idx += 1
        end_idx = self._sparse_values.indices.size
        self._last_node = self._node_indexes.size
        if self._sparse_values.values[-1] == 0:
            self._last_node = np.searchsorted(self._node_indexes, self._sparse_values.indices[-1])
            end_idx -= 1
        return self._sparse_values.indices[start_idx:end_idx].reshape(
            (end_idx-start_idx)//2, 2)

    def __get_holes(self):
        start_idx = 0
        if self._sparse_values.values[0] != 0:
            start_idx += 1
        indices = self._sparse_values.indices[start_idx:]
        if self._sparse_values.values[-1] == 0:
            indices = np.r_[indices, self._graph.node_indexes[-1]]
        return indices.astype("int").reshape(
            (indices.size)//2, 2)

    def build_sparse_values(self, holes):
        indices = np.sort(holes.ravel())
        values = np.arange(indices.size) % 2
        if (not indices.size) or indices[0] != 0:
            indices = np.r_[0, indices]
            values = np.r_[1, values]
        if self._sparse_values.values[-1] == 0:
            indices = np.r_[indices, self._sparse_values.indices[-1]]
            values = np.r_[values, 0]
        new_pileup = SparseValues(indices.astype("int"),
                            values.astype("bool"), sanitize=True)
        new_pileup.track_size = self._sparse_values.track_size
        return new_pileup

    def get_node_ids(self):
        node_ids = np.empty_like(self._holes)
        node_ids[:, 0] = np.digitize(self._holes[:, 0], self._node_indexes)
        node_ids[:, 1] = np.digitize(self._holes[:, 1], self._node_indexes, True)
        return node_ids

    def get_big_holes(self, internal_holes):
        return (internal_holes[:, 1]-internal_holes[0]) > self._max_size

    def _border_clean(self, full_mask, start_mask, end_mask, border_mask):
        full_starts = border_mask & start_mask
        full_ends = border_mask & end_mask
        start_mask -= full_starts
        end_mask -= full_starts
        return full_starts, full_ends

    def _handle_internal_holes(self, mask):
        holes = self._holes[mask]
        node_ids = self._node_ids[mask]
        is_start = holes[:, 0] == self._node_indexes[node_ids[:, 0]-1]
        true_internals = holes[~is_start]

    def _get_starts(self, pos, node_id):
        size = pos-self._node_indexes[node_id-1]
        return np.vstack((node_id, size))

    def _get_ends(self, pos, node_id):
        size = self._node_indexes[node_id]-pos
        return np.vstack((node_id, size))

    def _get_fulls(self, fulls):
        return np.vstack((
            fulls,
            self._node_indexes[fulls]-self._node_indexes[fulls-1]))

    def _handle_internal(self, internal_holes):
        keep = (internal_holes[:, 1]-internal_holes[:, 0]) > self._max_size
        self._kept_internals = internal_holes[keep]

    def classify_holes(self, hole_starts, hole_ends, start_ids, end_ids, is_multinodes):
        is_starts = hole_starts == self._node_indexes[start_ids-1]
        is_ends = hole_ends == self._node_indexes[end_ids]
        full_start_filter = is_starts & (is_ends | is_multinodes)
        full_end_filter = is_ends & (is_multinodes)
        starts_filter = is_starts ^ full_start_filter
        ends_filter = is_ends ^ (is_starts | is_multinodes)
        return starts_filter, ends_filter, full_start_filter, full_end_filter

    def _filter_touched_nodes(self, node_values):
        if not self._touched_nodes:
            return node_values
        touched = [i for i, node in enumerate(node_values[0])
                   if node+self._graph.min_node-1 in self._touched_nodes]
        self._not_touched = np.array(
            [i for i in node_values[0] if i+self._graph.min_node-1 not in self._touched_nodes],
            dtype="int")
        return node_values[:, touched]

    def run(self):
        analyzer = SegmentAnalyzer(
            self._holes, self._node_ids, self._node_indexes)
        analyzer.run()
        self._handle_internal(analyzer.internals)
        starts = analyzer.get_ends()
        fulls = self._filter_touched_nodes(analyzer.get_fulls())
        ends = analyzer.get_starts()
        linegraph = DividedLinegraph(starts, fulls, ends, self._graph,
                                     self._last_node, self._touched_nodes)
        mask = linegraph.filter_small(self._max_size)
        self._kept_borders = linegraph.get_masked(mask)
        return self.build_sparse_values(self.build_kept_holes())

    def build_kept_holes(self):
        starts, fulls, ends = self._kept_borders
        fulls = np.r_[fulls, self._not_touched]
        n_starts, n_fulls, n_ends = starts.shape[1], fulls.size, ends.shape[1]
        logging.debug("# %s, %s, %s", n_starts, n_fulls, n_ends)
        n_internals = self._kept_internals.shape[0]
        all_holes = np.empty((n_starts+n_fulls+n_ends+n_internals, 2),
                             dtype="int")
        if all_holes.shape == self._kept_internals.shape:
            return self._kept_internals
        if n_starts:
            all_holes[:n_starts, 0] = self._node_indexes[starts[0]] - starts[1]
            all_holes[:n_starts, 1] = self._node_indexes[starts[0]]
        n_border = n_starts+n_fulls+n_ends
        if fulls.size:
            all_holes[n_starts:n_starts+n_fulls, 0] = self._node_indexes[fulls-1]
            all_holes[n_starts:n_starts+n_fulls, 1] = self._node_indexes[fulls]
        if ends.size:
            all_holes[n_starts+n_fulls:n_border, 0] = self._node_indexes[ends[0]-1]
            all_holes[n_starts+n_fulls:n_border, 1] = self._node_indexes[ends[0]-1] + ends[1]
        all_holes[n_border:] = self._kept_internals
        all_holes.sort(axis=0)
        return all_holes

    def handle_border_holes(self, holes, node_ids):
        # node_offsets = self._node_indexes[node_ids]
        node_id_diffs = node_ids[:, 1]-node_ids[:, 0]
        full_nodes = (chain(range(i[0]+1, i[1])
                            for i in holes[node_id_diffs > 1]))
        full_nodes = [node for node in full_nodes if
                      self._graph.node_size(node) <= self._max_size]

        starts = np.vstack((holes[:, 0], self._node_indexes[node_ids[:, 1]+1]))
        ends = np.vstack((self._node_indexes[node_ids[:, 1]+1], holes[:, 1]))

    def find_border_holes(self):
        holes = self._holes.copy()
        holes[:, 1] += 1
        borders = np.digitize(self._graph.node_indexes, holes.ravel)
        border_holes = np.unique(borders)
        border_holes = border_holes[border_holes % 2 == 1]//2
        return border_holes


def test_holes_cleaner():
    import offsetbasedgraph as obg

    class S:
        indices = np.arange(0, 1000, 120)
        values = np.array([i % 2 for i in indices])

    class G:
        node_indexes = np.arange(0, 1001, 100)
        node_size = lambda x: 100

    graph = obg.Graph({i+1: obg.Block(100) for i in range(10)},
                      {i: [i+1] for i in range(0, 9)})
    graph.node_indexes = np.arange(0, 1001, 100)

if __name__ == "__main__":
    test_holes_cleaner()

    exit()


    starts = [[1, 2, 3], [10, 100, 3]]
    fulls = [[11, 12, 13, 14, 15], [10, 20, 30, 1, 1]]
    ends = [[21, 22], [60, 1]]
    tmp = {
        1: [11, 12],
        11: [21],
        12: [21],
        13: [22]
    }
    nodes = {i: obg.Block(10) for i in range(5, 100)}
    graph = obg.GraphWithReversals(nodes, tmp)
    l = LineGraph(np.array(starts), np.array(fulls), np.array(ends), graph)
    l.filter_small(20)




from .linear_filter import LinearFilter
from .shiftestimation import Opt, Treatment, PeakModel
import logging


class MultiGraphShiftEstimator(object):
    def __init__(self, position_tuples, genome_size):
        self._position_tuples = position_tuples
        self.genome_size = genome_size

    def get_estimates(self):
        opt = Opt()
        opt.gsize = self.genome_size
        treatment = Treatment(self._position_tuples)
        peakmodel = PeakModel(opt, treatment)
        peakmodel.build()
        return round(peakmodel.d)

    def to_linear_bed_file(self, file_name, read_length):
        f = open(file_name, "w")
        for direction in self._position_tuples.keys():
            for chrom, positions in self._position_tuples[direction].items():
                for pos in positions:
                    #print("Writing position %d" % pos)

                    start = pos
                    end = pos + read_length

                    if direction == "-":
                        start = pos - read_length
                        end = pos

                    f.writelines(["chr%s\t%d\t%d\t.\t0\t%s\n" % \
                                  (chrom, start, end, direction)])
        f.close()


    @classmethod
    def from_files(cls, chrom_names, graph_file_names, interval_json_file_names):

        start_positions = {
            "+": {name: [] for name in chrom_names},
            "-": {name: [] for name in chrom_names}
        }
        genome_size = 0

        for name, graph, intervals in zip(chrom_names,
                                          graph_file_names,
                                          interval_json_file_names):

            linear_filter = LinearFilter.from_vg_json_reads_and_graph(
                intervals, graph)

            positions = linear_filter.find_start_positions()
            start_positions["+"][name] = positions["+"]
            start_positions["-"][name] = positions["-"]

            genome_size += linear_filter._indexed_interval.length()

        logging.info("Using genome size %d" % genome_size)
        return cls(start_positions, genome_size)



import numpy as np
import logging
import offsetbasedgraph as obg
from pyvg.sequences import SequenceRetriever
from . import CallPeaksFromQvalues
from .peakscomparer import PeaksComparerV2, AnalysisResults
from .manually_classified_peaks import CheckOverlapWithManuallyClassifiedPeaks
from .analyse_peaks import LinearRegion
from .analysis.differentialbinding import main
from .analysis.fimowrapper import FimoFile
from .peakcollection import PeakCollection
from .nongraphpeaks import NonGraphPeakCollection


def analyse_peaks_whole_genome(args):
    chromosomes = args.chromosomes.split(",")
    results = AnalysisResults()
    for chrom in chromosomes:
        graph = obg.GraphWithReversals.from_numpy_file(
            args.graphs_dir + chrom + ".nobg")
        logging.info("Reading linear path")
        linear_path = obg.NumpyIndexedInterval.from_file(
            args.graphs_dir + chrom + "_linear_pathv2.interval")

        analyser = PeaksComparerV2(
            graph,
            args.results_dir + "macs_sequences_chr%s.fasta" % chrom,
            args.results_dir + "%s_sequences.fasta" % chrom,
            args.results_dir + "/fimo_macs_chr%s/fimo.txt" % chrom,
            args.results_dir + "/fimo_graph_chr%s/fimo.txt" % chrom,
            linear_path
        )
        results = results + analyser.results

    print(" === Final results for all chromosomes ===")
    print(results)

    results.to_file(args.out_file + ".pickle")
    with open(args.out_file + ".txt", "w") as f:
        f.write(str(results))
    logging.info("Wrote results as pickle to %s and as text to %s"
                 % (args.out_file + ".pickle", args.out_file + ".txt"))


def analyse_manually_classified_peaks(args):
    for chromosome in args.chromosomes.split():
        graph_file_name = args.graphs_location + "/" + chromosome + ".nobg"
        graph = obg.GraphWithReversals.from_numpy_file(graph_file_name)
        CheckOverlapWithManuallyClassifiedPeaks.from_graph_peaks_in_fasta(
                graph,
                args.graphs_location + "/" + chromosome + ".json",
                chromosome,
                args.reads_base_name + chromosome + "_sequences.fasta",
                args.regions_file,
                args.manually_classified_peaks_file)


def analyse_peaks(args):
    graph = obg.GraphWithReversals.from_numpy_file(args.ob_graph_file_name)

    end = int(args.graph_end)
    if end == 0:
        region = None
    else:
        region = LinearRegion(args.graph_chromosome,
                              int(args.graph_start), end)
    linear_path = obg.NumpyIndexedInterval.from_file(
        args.linear_path_name)
    PeaksComparerV2(
        graph,
        args.linear_peaks_fasta_file_name,
        args.graph_peaks_fasta_file_name,
        args.linear_peaks_fimo_results_file,
        args.graph_peaks_fimo_results_file,
        linear_path,
        region=region)


def differential_expression(args):

    test_name = args.test_name
    fimo_file_name = "fimo_%s_sequences/fimo.txt" % test_name
    peaks_file_name = "%s_max_paths.intervalcollection" % test_name
    subgraphs_file_name = "%s_sub_graphs.graphs.npz" % test_name
    node_ids_file_name = "%s_sub_graphs.nodeids.npz" % test_name
    graph = obg.GraphWithReversals.from_numpy_file(args.graph_name)
    res = main(
        FimoFile.from_file(fimo_file_name),
        PeakCollection.from_file(peaks_file_name, True),
        np.load(subgraphs_file_name),
        np.load(node_ids_file_name),
        graph)
    retriever = SequenceRetriever.from_vg_json_graph(args.vg_graph_name)
    out_f = open(test_name + "_diffexpr.fasta", "w")
    for expr_diff in res:
        main_seq = retriever.get_interval_sequence(expr_diff.main_path)
        out_f.write("> %s %s\n" % (expr_diff.peak_id, expr_diff.main_count))
        out_f.write(main_seq + "\n")
        var_seq = retriever.get_interval_sequence(expr_diff.var_path)
        out_f.write("> %sAlt %s\n" % (expr_diff.peak_id, expr_diff.var_count))
        out_f.write(var_seq + "\n")
    out_f.close()

def plot_motif_enrichment(args):
    from graph_peak_caller.motifenrichment import plot_true_positives
    fasta1 = args.fasta1
    fasta2 = args.fasta2
    meme = args.meme_motif_file

    plot_true_positives(
        [
            ("Graph Peak Caller", fasta1),
            ("MACS2", fasta2)
        ],
        meme,
        plot_title=args.plot_title,
        save_to_file=args.out_figure_file_name,
        run_fimo=args.run_fimo == "True"
    )


def intervals_to_fasta(args):
    logging.info("Getting sequence retriever")
    retriever = SequenceRetriever.from_vg_graph(args.vg_graph_file_name)
    logging.info("Getting intervals")
    intervals = obg.IntervalCollection.create_generator_from_file(
        args.intervals_file_name)
    logging.info("Writing to fasta")
    CallPeaksFromQvalues.intervals_to_fasta_file(
        intervals, args.out_file_name, retriever)


def linear_peaks_to_fasta(args):
    collection = NonGraphPeakCollection.from_bed_file(
        args.linear_reads_file_name)
    collection.set_peak_sequences_using_fasta(
        fasta_file_location=args.fasta_file)
    collection.save_to_sorted_fasta(args.out_file_name)
    logging.info("Saved sequences to %s" % args.out_file_name)



from .densepileup import DensePileup
import numpy as np
from .extender import Extender
import offsetbasedgraph as obg
import logging


class DagHoleCleaner(object):
    """
    Efficient Hole cleaner for DensePileup on Directed Acyclic Graph
    """

    def __init__(self, dense_pileup, hole_size):
        assert isinstance(dense_pileup, DensePileup)
        self.pileup = dense_pileup
        self._graph = dense_pileup.graph
        self.hole_size = hole_size

    def run(self):
        positions = self.get_left_side_of_holes()
        self.expand_hole_sides_to_right(positions)
        return self.pileup

    def get_left_side_of_holes(self):
        """
        Find all internal left sides by checking diff == -1
         and ignoring beginning of nodes
        Add all left hole side that are at beginning of node by
        explicitly checking all node starts
        """
        logging.info("Getting left side of holes")
        node_lengths = [self.pileup.data.node_size(node)
                        for node in self.pileup.data._nodes[0:-1]]
        start_node_indices = np.cumsum(node_lengths)
        diffs = np.diff(self.pileup.data._values * 1)

        if len(start_node_indices) > 0:
            diffs[start_node_indices-1] = 0  # Ignore start of nodes here, we cannot be sure they are holes

        left_hole_indices = np.where(diffs == -1)[0] + 1
        positions = self.pileup.data.value_indexes_to_nodes_and_offsets(
            left_hole_indices)

        # Check all node starts
        logging.info("Adding node starts")
        is_hole = set(
            np.where(self.pileup.data._values == False)[0])
        logging.info("Step 1 done")
        is_hole_and_start = is_hole.intersection(set(start_node_indices))
        logging.info("Step 2 done")
        is_hole_and_start = sorted(list(is_hole_and_start))
        logging.info("Step 3 done")
        start_positions = self.pileup.data.value_indexes_to_nodes_and_offsets(
            is_hole_and_start)

        logging.info("Going through each start")
        for start_position in start_positions:
            node = start_position[0]
            is_start_of_hole = True
            if len(self._graph.reverse_adj_list[-node]) == 0:
                continue  # Start of grap, not hole

            for in_node in self._graph.reverse_adj_list[-node]:
                values = self.pileup.data.values(-in_node)
                if not values[-1]:
                    is_start_of_hole = False
                    break

            if is_start_of_hole:
                positions.append(start_position)

        return positions

    def expand_hole_sides_to_right(self, positions):

        logging.info("Number of holes to expand: %d" % len(positions))
        extender = Extender(self._graph, self.hole_size)

        i = 0
        for position in positions:
            if i % 500 == 0:
                logging.info("Expanding hole %d/%d" % (i, len(positions)))
            i += 1

            node = position[0]
            offset = position[1]
            interval = obg.DirectedInterval(
                int(offset), int(offset) + 1, [node])
            areas = extender.extend_interval(interval)
            self.pileup.set_area_to_value(areas, True)

        self.pileup.threshold(0.5)



class CollectionIO(object):
    _obj_type = None

    def __init__(self, collection):
        self._collection = collection

    @classmethod
    def from_file(cls, file_name, graph):
        with open(file_name) as f:
            collection = [cls._obj_type.from_file_line(line, graph)
                          for line in f.readlines()]
        return cls(collection)

    def to_file(self, file_name):
        with open(file_name, "w") as f:
            f.writelines(item.to_file_line() for item in self._collection)
